from starfish.image import ImageStack
from starfish.pipeline.algorithmbase import AlgorithmBase


class RegistrationAlgorithmBase(AlgorithmBase):
    def run(self, stack) -> ImageStack:
        """Performs registration on the stack provided."""
        raise NotImplementedError()



from typing import Any, Dict, Iterable

import numpy as np
import pandas as pd
import regional


def swap(img):
    img_swap = img.swapaxes(0, img.ndim - 1)
    return img_swap


def stack_to_list(stack):
    num_ims = stack.shape[0]
    return [stack[im, :] for im in range(num_ims)]


def list_to_stack(list):
    return np.array(list)


def max_proj(stack):
    im = np.max(stack, axis=0)
    return im


def dataframe_to_multiindex(dataframe: pd.DataFrame) -> pd.MultiIndex:
    """Convert data in a DataFrame to a MultiIndex"""
    names, arrays = zip(*(dataframe.items()))
    return pd.MultiIndex.from_arrays(arrays=arrays, names=names)


def scale(stack, metric, clip=False):
    from starfish.stats import stack_describe
    stats = stack_describe(stack)
    ims = stack_to_list(stack)
    res = [im / s[metric] for im, s in zip(ims, stats)]
    return list_to_stack(res)


def melt(df: pd.DataFrame, new_index_name: Any, new_value_name: Any, melt_columns: Iterable) -> pd.DataFrame:
    """Melt all columns in `melt_columns` into tidy format, tiling all unspecified columns as identifiers

    Columns in `melt_columns` are aggregated into a new column. Their identifiers are stored together in a column
    whose index is `new_index_name`, and their values are stored together in a column whose name is `new_value_name`

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame to melt
    new_index_name : Any
        The name of the index column that will be created (typically a str)
    new_value_name : Any
        The name of the column storing the melted values (typically a str)
    melt_columns : Iterable
        Names of columns to be melted into tidy format.

    Examples
    --------
    >>> import pandas as pd
    >>> test = pd.DataFrame(
    ...     data = [
    ...         ['x', '1', '4'],
    ...         ['y', '2', '5'],
    ...         ['z', '3', '6']
    ...     ],
    ...     columns = ['id', 'target_1', 'target_2']
    ... )
    >>> melt(test, 'orig_col', 'melted_values', ['target_1', 'target_2'])
        id  orig_col    melted_values
    0   x   target_1    a
    1   y   target_1    b
    2   z   target_1    c
    3   x   target_2    d
    4   y   target_2    e
    5   z   target_2    f

    Returns
    -------
    pd.DataFrame :
        Melted (tidy) DataFrame

    """
    # any value not passed in melt_columns is retained as an id_var
    id_vars = [col for col in df.columns if col not in melt_columns]

    melted = pd.melt(
        df,
        id_vars=id_vars,
        value_vars=melt_columns,
        var_name=new_index_name,
        value_name=new_value_name
    )

    # melt does not know the dtypes of the new columns, but pandas can normally guess them
    melted = melted.infer_objects()

    return melted


def spots_to_geojson(spots_viz):
    '''
    Convert spot geometrical data to geojson format
    '''

    def make_dict(row):
        row = row[1]
        d = dict()
        d['properties'] = {'id': int(row.spot_id), 'radius': int(row.r)}
        d['geometry'] = {'type': 'Point', 'coordinates': [int(row.x), int(row.y)]}
        return d

    return [make_dict(row) for row in spots_viz.iterrows()]


def geojson_to_region(geojson: Dict[Any, Any]) -> regional.many:
    """
    Convert geojson data to region geometrical data.
    """
    def make_region(geometry):
        assert geometry['geometry']['type'] == "Polygon"

        return regional.one([(coordinates[0], coordinates[1]) for coordinates in geometry['geometry']['coordinates']])

    return regional.many([make_region(geometry) for geometry in geojson])


def relabel(image):
    '''
    This is a local implementation of centrosome.cpmorphology.relabel
    to remove this dependency from starfish.

    Takes a labelled image and relabels each image object consecutively.
    Original code from:
    https://github.com/CellProfiler/centrosome/blob/master/centrosome/cpmorphology.py

    They use a BSD-3 license, which would then have to be propagated to starfish,
    this could be an issue.

    Args
    ----
    image: numpy.ndarray
        A 2d integer array representation of an image with labels

    Returns
    -------
    new_image: numpy.ndarray
        A 2d integer array representation of an image wiht new labels

    n_labels: int
        The number of new unique labels
    '''

    # I've set this as a separate function, rather than binding it to the
    # _WatershedSegmenter object for now

    unique_labels = set(image[image != 0])
    n_labels = len(unique_labels)

    # if the image is unlabelled, return original image
    # warning/message required?
    if n_labels == 0:
        return (image, 0)

    consec_labels = np.arange(n_labels) + 1
    lab_table = np.zeros(max(unique_labels) + 1, int)
    lab_table[[x for x in unique_labels]] = consec_labels

    # Use the label table to remap all of the labels
    new_image = lab_table[image]

    return new_image, n_labels



from typing import Tuple, Union

import numpy as np
from scipy.ndimage import fourier_shift
from skimage.feature import register_translation

from starfish.constants import Indices
from starfish.image import ImageStack
from starfish.util.argparse import FsExistsType
from ._base import RegistrationAlgorithmBase


class FourierShiftRegistration(RegistrationAlgorithmBase):
    """
    Implements fourier shift registration.  TODO: (dganguli) FILL IN DETAILS HERE PLS.

    Performs a simple translation registration.

    See Also
    --------
    https://en.wikipedia.org/wiki/Phase_correlation
    """
    def __init__(self, upsampling: int, reference_stack: Union[str, ImageStack], **kwargs) -> None:
        self.upsampling = upsampling
        if isinstance(reference_stack, ImageStack):
            self.reference_stack = reference_stack
        else:
            self.reference_stack = ImageStack.from_path_or_url(reference_stack)

    @classmethod
    def add_arguments(cls, group_parser) -> None:
        group_parser.add_argument("--upsampling", default=1, type=int, help="Amount of up-sampling")
        group_parser.add_argument(
            "--reference-stack", type=FsExistsType(), required=True,
            help="The image stack to align the input image stack to.")

    def run(self, image: ImageStack) -> ImageStack:
        # TODO: (ambrosejcarr) is this the appropriate way of dealing with Z in registration?
        mp = image.max_proj(Indices.CH, Indices.Z)
        reference_image = self.reference_stack.max_proj(Indices.ROUND, Indices.CH, Indices.Z)

        for r in range(image.num_rounds):
            # compute shift between maximum projection (across channels) and dots, for each round
            # TODO: make the max projection array ignorant of axes ordering.
            shift, error = compute_shift(mp[r, :, :], reference_image, self.upsampling)
            print(f"For round: {r}, Shift: {shift}, Error: {error}")

            for c in range(image.num_chs):
                for z in range(image.num_zlayers):
                    # apply shift to all zlayers, channels, and imaging rounds
                    indices = {Indices.ROUND: r, Indices.CH: c, Indices.Z: z}
                    data, axes = image.get_slice(indices=indices)
                    assert len(axes) == 0
                    result = shift_im(data, shift)
                    image.set_slice(indices=indices, data=result)

        return image


def compute_shift(
        im: np.ndarray, ref: np.ndarray, upsample_factor: int=1
) -> Tuple[np.ndarray, float]:
    """calculate subpixel image translation through cross-correlation

    Parameters
    ----------
    im : np.ndarray
        reference image
    ref : np.ndarray
        target image
    upsample_factor : int
        images are registered to within 1 / upsample_factor of a pixel

    Returns
    -------
    np.ndarray :
        shift vector required to register ref
    float :
        translation invariant normalized RMS error
    """
    shift, error, _ = register_translation(im, ref, upsample_factor)
    return shift, error


def shift_im(im: np.ndarray, shift: np.ndarray) -> np.ndarray:
    """register image according to the provided shift values"""
    fim_shift = fourier_shift(np.fft.fftn(im), shift * -1)
    im_shift = np.fft.ifftn(fim_shift)
    return im_shift.real



from starfish.pipeline.features.spots.detector.gaussian import GaussianSpotDetector
from starfish.pipeline.filter.white_tophat import WhiteTophat
from starfish.pipeline.registration.fourier_shift import FourierShiftRegistration
from starfish.util.synthesize import SyntheticData
import numpy as np
from starfish.constants import Indices
from starfish.image import ImageStack


def test_iss_pipeline():
    np.random.seed(2)
    synthesizer = SyntheticData(n_spots=5)
    codebook = synthesizer.codebook()
    true_intensities = synthesizer.intensities(codebook=codebook)
    image = synthesizer.spots(intensities=true_intensities)

    dots_data = image.max_proj(Indices.ROUND, Indices.CH, Indices.Z)
    dots = ImageStack.from_numpy_array(dots_data.reshape((1, 1, 1, *dots_data.shape)))

    wth = WhiteTophat(masking_radius=15)
    wth.run(image)
    wth.run(dots)

    fsr = FourierShiftRegistration(upsampling=1000, reference_stack=dots)
    fsr.run(image)

    min_sigma = 1.5
    max_sigma = 5
    num_sigma = 10
    threshold = 1e-4
    gsd = GaussianSpotDetector(
        min_sigma=min_sigma,
        max_sigma=max_sigma,
        num_sigma=num_sigma,
        threshold=threshold,
        measurement_type='max',
    )
    blobs_image = dots.numpy_array.reshape(1, *dots_data.shape)
    intensities = gsd.find(data_stack=image, blobs_image=blobs_image)
    assert intensities.shape[0] == 5

    codebook.metric_decode(intensities, max_distance=1, min_intensity=0, norm_order=2)



from typing import Optional

import numpy as np
from scipy.ndimage.filters import maximum_filter, minimum_filter
from skimage.morphology import disk, ball

from starfish.image import ImageStack
from ._base import FilterAlgorithmBase


class WhiteTophat(FilterAlgorithmBase):
    """
    Performs "white top hat" filtering of an image to enhance spots. "White top hat filtering"
    finds spots that are both smaller and brighter than their surroundings.

    See Also
    --------
    https://en.wikipedia.org/wiki/Top-hat_transform
    """

    def __init__(self, masking_radius: int, is_volume: bool=False, **kwargs) -> None:
        """
        Instance of a white top hat morphological masking filter which masks objects larger
        than `masking_radius`

        Parameters
        ----------
        masking_radius : int
            radius of the morphological masking structure in pixels
        is_volume : int
            If True, 3d (z, y, x) volumes will be filtered, otherwise, filter 2d tiles
            independently.

        """
        self.masking_radius = masking_radius
        self.is_volume = is_volume

    @classmethod
    def add_arguments(cls, group_parser) -> None:
        group_parser.add_argument(
            "--masking-radius", default=15, type=int,
            help="diameter of morphological masking disk in pixels")

    # TODO dganguli: any reason we're not using the white_tophat method?
    # https://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.ndimage.\
    # morphology.white_tophat.html
    def white_tophat(self, image: np.ndarray) -> np.ndarray:
        """
        run a white top hat morphological masking filter which detects peaks that are smaller
        and brighter than their surroundings

        Parameters
        ----------
        image: np.ndarray
            2 or 3-d image

        Returns
        -------
        np.ndarray:
            filtered image

        """

        if image.dtype.kind != "u":
            raise TypeError("images should be stored in an unsigned integer array")

        if self.is_volume:
            structuring_element = ball(self.masking_radius)
        else:
            structuring_element = disk(self.masking_radius)

        min_filtered = minimum_filter(image, footprint=structuring_element)
        max_filtered = maximum_filter(min_filtered, footprint=structuring_element)
        filtered_image = image - np.minimum(image, max_filtered)
        return filtered_image

    def run(
            self, stack: ImageStack, in_place: bool=True, verbose: bool=False) \
            -> Optional[ImageStack]:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            If True, report on the percentage completed (default = False) during processing

        Returns
        -------
        Optional[ImageStack] :
            if in-place is False, return the results of filter as a new stack

        """
        result = stack.apply(
            self.white_tophat, is_volume=self.is_volume, verbose=verbose, in_place=in_place)
        if not in_place:
            return result
        return None



import json

import numpy as np
import pandas as pd

from ._validated_table import ValidatedTable


class SpotAttributes(ValidatedTable):

    required_fields = {
        'x',  # spot x-coordinate
        'y',  # spot y-coordinate
        'radius',  # spot radius
        'intensity',  # intensity of spot (commonly max or average)
        'spot_id'  # integer spot id
    }

    def __init__(self, spot_attributes: pd.DataFrame) -> None:
        """Construct a SpotAttributes instance

        Parameters
        ----------
        spot_attributes : pd.DataFrame

        """
        super().__init__(spot_attributes, SpotAttributes.required_fields)

    def save_geojson(self, output_file_name: str) -> None:
        """Save to geojson for web visualization

        Parameters
        ----------
        output_file_name : str
            name of output json file

        """

        # TODO ambrosejcarr: write a test for this
        geojson = [
            {
                'properties': {'id': int(row.spot_id), 'radius': int(row.r)},
                'geometry': {'type': 'Point', 'coordinates': [int(row.x), int(row.y)]}
            } for index, row in self.data.iterrows()
        ]

        with open(output_file_name, 'w') as f:
            f.write(json.dumps(geojson))

    def display(self, background_image: np.ndarray) -> None:
        """

        Parameters
        ----------
        background_image : np.ndarray
            image on which to plot spots. If 3d, take a max projection

        Returns
        -------

        """
        # TODO ambrosejcarr: re-implement this from the show() method
        pass



from enum import Enum


class AugmentedEnum(Enum):
    def __hash__(self):
        return self.value.__hash__()

    def __eq__(self, other):
        if isinstance(other, type(self)) or isinstance(other, str):
            return self.value == other
        return False

    def __str__(self) -> str:
        return self.value


class Coordinates(AugmentedEnum):
    Z = 'z'
    Y = 'y'
    X = 'x'


class Indices(AugmentedEnum):
    ROUND = 'r'
    CH = 'c'
    Z = 'z'


class Features:
    """
    contains constants relating to the codebook and feature (spot/pixel) representations of the
    image data
    """

    AXIS = 'features'
    TARGET = 'target'
    CODEWORD = 'codeword'
    CODE_VALUE = 'v'
    SPOT_RADIUS = 'radius'
    DISTANCE = 'distance'
    Z = 'z'
    Y = 'y'
    X = 'x'



import collections
import os
import warnings
from copy import deepcopy
from functools import partial
from itertools import product
from typing import Any, Callable, Iterable, Iterator, List, Mapping, MutableSequence, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
import xarray as xr
from scipy.ndimage.filters import gaussian_filter
from scipy.stats import scoreatpercentile
from skimage import exposure
from slicedimage import Reader, Writer, TileSet, Tile
from slicedimage.io import resolve_path_or_url
from tqdm import tqdm

from starfish.constants import Coordinates, Indices, Features
from starfish.errors import DataFormatWarning
from starfish.intensity_table import IntensityTable
from starfish.pipeline.features.spot_attributes import SpotAttributes


_DimensionMetadata = collections.namedtuple("_DimensionMetadata", ['order', 'required'])


class ImageStack:
    """Container for a TileSet (field of view)

    Methods
    -------
    get_slice    retrieve a slice of the image tensor
    set_slice    set a slice of the image tensor
    apply        apply a 2d or 3d function across all Tiles in the image tensor
    max_proj     return a max projection over one or more axis of the image tensor
    show_stack   show an interactive, pageable view of the image tensor, or a slice of the image tensor
    write        save the (potentially modified) image tensor to disk

    Properties
    ----------
    num_chs      the number of channels stored in the image tensor
    num_rounds   the number of imaging rounds stored in the image tensor
    num_zlayers  the number of z-layers stored in the image tensor
    numpy_array  the 5-d image tensor is stored in this array
    raw_shape    the shape of the image tensor (in integers)
    shape        the shape of the image tensor by categorical index (channels, imaging rounds, z-layers)
    """

    AXES_DATA: Mapping[Indices, _DimensionMetadata] = {
        Indices.ROUND: _DimensionMetadata(0, True),
        Indices.CH: _DimensionMetadata(1, True),
        Indices.Z: _DimensionMetadata(2, False),
    }
    N_AXES = max(data.order for data in AXES_DATA.values()) + 1

    def __init__(self, image_partition):
        self._image_partition = image_partition
        self._tile_shape = image_partition.default_tile_shape

        # Examine the tiles to figure out the right kind (int, float, etc.) and size.  We require that all the tiles
        # have the same kind of data type, but we do not require that they all have the same size of data type.  The
        # allocated array is the highest size we encounter.
        kind = None
        max_size = 0
        for tile in self._image_partition.tiles():
            dtype = tile.numpy_array.dtype
            if kind is None:
                kind = dtype.kind
            else:
                if kind != dtype.kind:
                    raise TypeError("All tiles should have the same kind of dtype")
            if dtype.itemsize > max_size:
                max_size = dtype.itemsize
            if self._tile_shape is None:
                self._tile_shape = tile.tile_shape
            elif tile.tile_shape is not None and self._tile_shape != tile.tile_shape:
                raise ValueError("Starfish does not support tiles that are not identical in shape")

        shape: MutableSequence[int] = []
        dims: MutableSequence[str] = []
        for ix in range(ImageStack.N_AXES):
            size_for_axis: Optional[int] = None
            dim_for_axis: Optional[Indices] = None

            for axis_name, axis_data in ImageStack.AXES_DATA.items():
                if ix == axis_data.order:
                    size_for_axis = self._get_dimension_size(axis_name)
                    dim_for_axis = axis_name

            if size_for_axis is None or dim_for_axis is None:
                raise ValueError(
                    f"Could not find entry for the {ix}th axis in ImageStack.AXES_DATA")

            shape.append(size_for_axis)
            dims.append(dim_for_axis.value)

        shape.extend(self._tile_shape)
        dims.extend([Coordinates.Y.value, Coordinates.X.value])

        # now that we know the tile data type (kind and size), we can allocate the data array.
        self._data = xr.DataArray(
            np.zeros(
                shape=shape,
                dtype=np.dtype(f"{kind}{max_size}"),
            ),
            dims=dims,
        )

        # iterate through the tiles and set the data.
        for tile in self._image_partition.tiles():
            h = tile.indices[Indices.ROUND]
            c = tile.indices[Indices.CH]
            zlayer = tile.indices.get(Indices.Z, 0)
            data = tile.numpy_array
            if max_size != data.dtype.itemsize:
                if data.dtype.kind == "i" or data.dtype.kind == "u":
                    # fixed point can be done with a simple multiply.
                    src_range = np.iinfo(data.dtype).max - np.iinfo(data.dtype).min + 1
                    dst_range = np.iinfo(self._data.dtype).max - np.iinfo(self._data.dtype).min + 1
                    data = data * (dst_range / src_range)
                warnings.warn(
                    f"Tile "
                    f"(R: {tile.indices[Indices.ROUND]} C: {tile.indices[Indices.CH]} "
                    f"Z: {tile.indices[Indices.Z]}) has "
                    f"dtype {data.dtype}.  One or more tiles is of a larger dtype {self._data.dtype}.",
                    DataFormatWarning)
            self.set_slice(indices={Indices.ROUND: h, Indices.CH: c, Indices.Z: zlayer}, data=data)
        # set_slice will mark the data as needing writeback, so we need to unset that.
        self._data_needs_writeback = False

    @classmethod
    def from_url(cls, url: str, baseurl: Optional[str]):
        """
        Constructs an ImageStack object from a URL and a base URL.

        The following examples will all load from the same location:
          1. url: https://www.example.com/images/hybridization.json  baseurl: None
          2. url: https://www.example.com/images/hybridization.json  baseurl: I_am_ignored
          3. url: hybridization.json  baseurl: https://www.example.com/images
          4. url: images/hybridization.json  baseurl: https://www.example.com

        Parameters:
        -----------
        url : str
            Either an absolute URL or a relative URL referring to the image to be read.
        baseurl : Optional[str]
            If url is a relative URL, then this must be provided.  If url is an absolute URL, then this parameter is
            ignored.
        """
        image_partition = Reader.parse_doc(url, baseurl)

        return cls(image_partition)

    @classmethod
    def from_path_or_url(cls, url_or_path: str) -> "ImageStack":
        """
        Constructs an ImageStack object from an absolute URL or a filesystem path.

        The following examples will all load from the same location:
          1. url_or_path: file:///Users/starfish-user/images/hybridization.json
          2. url_or_path: /Users/starfish-user/images/hybridization.json

        Parameters:
        -----------
        url_or_path : str
            Either an absolute URL or a filesystem path to an imagestack.
        """
        _, relativeurl, baseurl = resolve_path_or_url(url_or_path)
        return cls.from_url(relativeurl, baseurl)

    @property
    def numpy_array(self):
        """Retrieves the image data as a numpy array."""
        result = self._data.values
        return result

    @numpy_array.setter
    def numpy_array(self, data):
        """Sets the image's data from a numpy array."""
        self._data.values = data.view()
        self._data_needs_writeback = True

    def get_slice(
            self,
            indices: Mapping[Indices, Union[int, slice]]
    ) -> Tuple[np.ndarray, Sequence[Indices]]:
        """
        Given a dictionary mapping the index name to either a value or a slice range, return a numpy array representing
        the slice, and a list of the remaining axes beyond the normal x-y tile.

        Example:
            ImageStack axes: H, C, and Z with shape 3, 4, 5, respectively.
            ImageStack Implicit axes: X, Y with shape 10, 20, respectively.
            Called to slice with indices {Z: 5}.
            Result: a 4-dimensional numpy array with shape (3, 4, 20, 10) and the remaining axes [H, C].

        Example:
            Original axes: H, C, and Z.
            Implicit axes: X, Y.
            Called to slice with indices {Z: 5, C: slice(2, 4)}.
            Result: a 4-dimensional numpy array with shape (3, 2, 20, 10) and the remaining axes [H, C].
        """
        slice_list, axes = self._build_slice_list(indices)
        result = self._data.values[slice_list]
        return result, axes

    def set_slice(
            self,
            indices: Mapping[Indices, Union[int, slice]],
            data: np.ndarray,
            axes: Sequence[Indices]=None):
        """
        Given a dictionary mapping the index name to either a value or a slice range and a source numpy array, set the
        slice of the array of this ImageStack to the values in the source numpy array.  If the optional parameter axes
        is provided, that represents the axes of the numpy array beyond the x-y tile.

        Example:
            ImageStack axes: H, C, and Z with shape 3, 4, 5, respectively.
            ImageStack Implicit axes: X, Y with shape 10, 20, respectively.
            Called to set a slice with indices {Z: 5}.
            Data: a 4-dimensional numpy array with shape (3, 4, 10, 20)
            Result: Replace the data for Z=5.

        Example:
            ImageStack axes: H, C, and Z. (shape 3, 4, 5)
            ImageStack Implicit axes: X, Y. (shape 10, 20)
            Called to set a slice with indices {Z: 5, C: slice(2, 4)}.
            Data: a 4-dimensional numpy array with shape (3, 2, 10, 20)
            Result: Replace the data for Z=5, C=2-3.
        """
        slice_list, expected_axes = self._build_slice_list(indices)

        if axes is not None:
            if len(axes) != len(data.shape) - 2:
                raise ValueError("data shape ({}) should be the axes ({}) and (x,y).".format(data.shape, axes))
            move_src = list()
            move_dst = list()
            for src_idx, axis in enumerate(axes):
                try:
                    dst_idx = expected_axes.index(axis)
                except ValueError:
                    raise ValueError("Unexpected axis {}.  Expecting only {}.".format(axis, expected_axes))
                if src_idx != dst_idx:
                    move_src.append(src_idx)
                    move_dst.append(dst_idx)

            if len(move_src) != 0:
                data = data.view()
                np.moveaxis(data, move_src, move_dst)

        if self._data[slice_list].shape != data.shape:
            raise ValueError("source shape {} mismatches destination shape {}".format(
                data.shape, self._data[slice_list].shape))

        self._data.values[slice_list] = data
        self._data_needs_writeback = True

    # TODO ambrosejcarr: update to use IntensityTable instead of SpotAttributes
    def show_stack(
            self, indices: Mapping[Indices, Union[int, slice]],
            color_map: str= 'gray', figure_size: Tuple[int, int]=(10, 10),
            show_spots: Optional[SpotAttributes]=None,
            rescale: bool=False, p_min: Optional[float]=None, p_max: Optional[float]=None, **kwargs):
        """Create an interactive visualization of an image stack

        Produces a slider that flips through the selected volume tile-by-tile. Supports manual adjustment of dynamic
        range.

        Parameters
        ----------
        indices : Mapping[Indices, Union[int, slice]],
            Indices to select a volume to visualize. Passed to `Image.get_slice()`.
            See `Image.get_slice()` for examples.
        color_map : str (default = 'gray')
            string id of a matplotlib colormap
        figure_size : Tuple[int, int] (default = (10, 10))
            size of the figure in inches
        show_spots : Optional[SpotAttributes]
            [Preliminary functionality] if provided, should be a SpotAttribute table that corresponds
            to the volume being displayed. This will be paired automatically in the future.
        rescale : bool (default = False)
            if True, rescale the data to exclude high and low-value outliers (see skimage.exposure.rescale_intensity).
        p_min: float
            clip values below this intensity percentile. If provided, overrides rescale, above. (default = None)
        p_max: float
            clip values above this intensity percentile. If provided, overrides rescale, above. (default = None)

        Raises
        ------
        ValueError :
            User must select one of rescale or p_min/p_max to adjust the image dynamic range. If both are selected, a
            ValueError is raised.

        Notes
        -----
        For this widget to function interactively in the notebook, after ipywidgets has been installed, the user must
        register the widget with jupyter by typing the following command into the terminal:
        jupyter nbextension enable --py widgetsnbextension

        """

        from ipywidgets import interact, fixed
        import matplotlib.pyplot as plt

        if not indices:
            raise ValueError('indices may not be an empty dict or None')

        # get the requested chunk, linearize the remaining data into a sequence of tiles
        data, remaining_inds = self.get_slice(indices)

        # identify the dimensionality of data with all dimensions other than x, y linearized
        if len(data.shape) >= 3:
            n = np.product(data.shape[:-2])
        else:
            raise ValueError(
                f'a stack with dimensionality >= 3 is required, the provided indexer produced a '
                f'stack with shape {data.shape}')

        # linearize the array
        linear_view: np.ndarray = data.reshape((n,) + data.shape[-2:])

        # set the labels for the linearized tiles
        labels: List[List[str]] = []
        for index, size in zip(remaining_inds, data.shape[:-2]):
            labels.append([f'{index}{n}' for n in range(size)])

        # mypy thinks this has an incompatible type "Iterator[Tuple[Any, ...]]";
        # it expects "Iterable[List[str]]"
        labels = list(product(*labels))  # type: ignore

        n = linear_view.shape[0]

        if rescale and any((p_min, p_max)):
            raise ValueError('select one of rescale and p_min/p_max to rescale image, not both.')

        elif rescale:
            print("Rescaling ...")
            vmin, vmax = scoreatpercentile(data, (0.5, 99.5))
            linear_view = exposure.rescale_intensity(
                linear_view,
                in_range=(vmin, vmax),
                out_range=np.float32
            ).astype(np.float32)

        elif p_min or p_max:
            print("Clipping ...")
            a_min, a_max = scoreatpercentile(
                linear_view,
                (p_min if p_min else 0, p_max if p_max else 100)
            )
            linear_view = np.clip(linear_view, a_min=a_min, a_max=a_max)

        # Create the plot
        fig, ax = plt.subplots(figsize=figure_size)
        im = ax.imshow(linear_view[0], cmap=color_map)
        ax.set_xticks([])
        ax.set_yticks([])

        if show_spots is not None:
            circs, circle_mask = self._show_spots(result_df=show_spots, ax=ax, n_slices=linear_view.shape[0])

        def show_plane(ax, plane, plane_index, cmap="gray", title=None):
            # Update the image in the current plane
            im.set_data(plane)

            # Update the spots, if necessary
            if show_spots is not None:
                # TODO: vectorize/broadcast?
                for i, state in enumerate(circle_mask[plane_index]):
                    circs[i].set_visible(state)

            if title:
                ax.set_title(title)

        def display_slice(plane_index, ax):
            show_plane(ax, linear_view[plane_index], plane_index, title=f'{labels[plane_index]}', cmap=color_map)

        interact(display_slice, ax=fixed(ax), plane_index=(0, n - 1))

        return

    @staticmethod
    def _show_spots(result_df, ax, size=1, n_slices=0, z_dist=1, scale_radius=5):
        """function to plot spot finding results on top of any image as hollow red circles

        called spots are colored by category

        Parameters:
        -----------
        result_df : pd.Dataframe
            result dataframe containing spot calls that correspond to the image channel
        ax, matplotlib.Axes.Axis
            axis to plot spots on
        size :
            Line width for the displayed spots
        n_slices : int
            The number of z slices being displayed. Default: 0
        z_dist : float
            The max distance a spot can be from the displayed slice to be visible. Default (1.5)
        scale_radius : float
            The size of the displayed spots is the scale_radius multiplied by the fit radius. Default: 5


        """
        from matplotlib import pyplot as plt

        n_circles = result_df.shape[0]

        # Create the mask
        if n_slices != 0 and 'z' in result_df.columns:
            circle_mask = np.zeros((n_slices, n_circles), dtype='bool')

            for i in range(n_slices):
                on_circles = np.abs(result_df['z'] - i) < z_dist
                circle_mask[i] = on_circles
        else:
            circle_mask = np.ones((n_slices, n_circles), dtype='bool')

        # Plot the spots
        circs = []
        for i in range(n_circles):
            r, x, y = result_df.loc[i, ['r', 'x', 'y']]  # radius is a duplicate, and is present twice
            c = plt.Circle((y, x), r * scale_radius, color='r', linewidth=size, fill=False)
            circs.append(c)
            ax.add_patch(c)

        return circs, circle_mask

    @staticmethod
    def _build_slice_list(
            indices: Mapping[Indices, Union[int, slice]]
    ) -> Tuple[Tuple[Union[int, slice], ...], Sequence[Indices]]:
        slice_list: MutableSequence[Union[int, slice]] = [
            slice(None, None, None)
            for _ in range(ImageStack.N_AXES)
        ]
        axes = []
        removed_axes = set()
        for name, value in indices.items():
            idx = ImageStack.AXES_DATA[name].order
            if not isinstance(value, slice):
                removed_axes.add(name)
            slice_list[idx] = value

        for dimension_value, dimension_name in sorted([
            (dimension_value.order, dimension_name)
            for dimension_name, dimension_value in ImageStack.AXES_DATA.items()
        ]):
            if dimension_name not in removed_axes:
                axes.append(dimension_name)

        return tuple(slice_list), axes

    def _iter_indices(self, is_volume: bool=False) -> Iterator[Mapping[Indices, int]]:
        """Iterate over indices of image tiles or image volumes if is_volume is True

        Parameters
        ----------
        is_volume : bool
            If True, yield indices necessary to extract volumes from self, else return
            indices for tiles

        Yields
        ------
        Dict[str, int]
            Mapping of dimension name to index

        """
        for round_ in np.arange(self.shape[Indices.ROUND]):
            for ch in np.arange(self.shape[Indices.CH]):
                if is_volume:
                    yield {Indices.ROUND: round_, Indices.CH: ch}
                else:
                    for z in np.arange(self.shape[Indices.Z]):
                        yield {Indices.ROUND: round_, Indices.CH: ch, Indices.Z: z}

    def _iter_tiles(
            self, indices: Iterable[Mapping[Indices, Union[int, slice]]]
    ) -> Iterable[np.ndarray]:
        """Given an iterable of indices, return a generator of numpy arrays from self

        Parameters
        ----------
        indices, Iterable[Mapping[str, int]]
            Iterable of indices that map a dimension (str) to a value (int)

        Yields
        ------
        np.ndarray
            Numpy array that corresponds to provided indices
        """
        for inds in indices:
            array, axes = self.get_slice(inds)
            yield array

    def apply(self, func, is_volume=False, in_place=True, verbose: bool=False, **kwargs) -> "ImageStack":
        """Apply func over all tiles or volumes in self

        Parameters
        ----------
        func : Callable
            Function to apply. must expect a first argument which is a 2d or 3d numpy array (see is_volume) and return a
            np.ndarray. If inplace is True, must return an array of the same shape.
        is_volume : bool
            (default False) If True, pass 3d volumes (x, y, z) to func
        in_place : bool
            (default True) If True, function is executed in place. If n_proc is not 1, the tile or
            volume will be copied once during execution. If false, a new ImageStack object will be produced.
        verbose : bool
            If True, report on the percentage completed (default = False) during processing
        kwargs : dict
            Additional arguments to pass to func

        Returns
        -------
        ImageStack :
            If inplace is False, return a new ImageStack, otherwise return a reference to the original stack with
            data modified by application of func
        """
        if not in_place:
            image_stack = deepcopy(self)
            return image_stack.apply(func, is_volume=is_volume, in_place=True, verbose=verbose, **kwargs)

        mapfunc: Callable = map  # TODO: ambrosejcarr posix-compliant multiprocessing
        indices = list(self._iter_indices(is_volume=is_volume))

        if verbose:
            tiles = tqdm(self._iter_tiles(indices))
        else:
            tiles = self._iter_tiles(indices)

        applyfunc: Callable = partial(func, **kwargs)
        results = mapfunc(applyfunc, tiles)

        for r, inds in zip(results, indices):
            self.set_slice(inds, r)

        return self

    def transform(self, func, is_volume=False, verbose=False, **kwargs) -> List[Any]:
        """Apply func over all tiles or volumes in self

        Parameters
        ----------
        func : Callable
            Function to apply. must expect a first argument which is a 2d or 3d numpy array (see is_volume) but
            may return any object type
        is_volume : bool
            (default False) If True, pass 3d volumes (x, y, z) to func
        verbose : bool
            If True, report on the percentage completed (default = False) during processing
        kwargs : dict
            Additional arguments to pass to func being applied

        Returns
        -------
        List[Any] :
            The results of applying func to stored image data
        """
        mapfunc: Callable = map
        indices = list(self._iter_indices(is_volume=is_volume))

        if verbose:
            tiles = tqdm(self._iter_tiles(indices))
        else:
            tiles = self._iter_tiles(indices)

        applyfunc: Callable = partial(func, **kwargs)
        results = mapfunc(applyfunc, tiles)

        return list(zip(results, indices))

    @property
    def tile_metadata(self) -> pd.DataFrame:
        """return a table containing Tile metadata

        Returns
        -------
        pd.DataFrame :
            dataframe containing per-tile metadata information for each image. Guaranteed to include information on
            channel, imaging round, z_layer, and barcode index. Also contains any information stored in the
            extras field for each tile in hybridization.json

        """

        data: collections.defaultdict = collections.defaultdict(list)
        index_keys = set(
            key
            for tile in self._image_partition.tiles()
            for key in tile.indices.keys())
        extras_keys = set(
            key
            for tile in self._image_partition.tiles()
            for key in tile.extras.keys())
        duplicate_keys = index_keys.intersection(extras_keys)
        if len(duplicate_keys) > 0:
            duplicate_keys_str = ", ".join([str(key) for key in duplicate_keys])
            raise ValueError(
                f"keys ({duplicate_keys_str}) was found in both the Tile specification and extras field. Tile "
                f"specification keys may not be duplicated in the extras field.")

        for tile in self._image_partition.tiles():
            for k in index_keys:
                data[k].append(tile.indices.get(k, None))
            for k in extras_keys:
                data[k].append(tile.extras.get(k, None))

            if 'barcode_index' not in tile.extras:
                round_ = tile.indices[Indices.ROUND]
                ch = tile.indices[Indices.CH]
                z = tile.indices.get(Indices.Z, 0)
                barcode_index = (((z * self.num_rounds) + round_) * self.num_chs) + ch

                data['barcode_index'].append(barcode_index)

        return pd.DataFrame(data)

    @property
    def raw_shape(self) -> Tuple[int, int, int, int, int]:
        """
        Returns the shape of the 5-d image tensor stored as self.image

        Returns
        -------
        Tuple[int, int, int, int, int] :
            The size of the image tensor
        """
        return self._data.shape

    @property
    def shape(self) -> collections.OrderedDict:
        """
        Returns the shape of the space that this image inhabits.  It does not include the dimensions of the image
        itself.  For instance, if this is an X-Y image in a C-H-Y-X space, then the shape would include the dimensions C
        and H.

        Returns
        -------
        An ordered mapping between index names to the size of the index.
        """
        # TODO: (ttung) Note that the return type should be ..OrderedDict[Any, str], but python3.6 has a bug where this
        # breaks horribly.  Can't find a bug id to link to, but see
        # https://stackoverflow.com/questions/41207128/how-do-i-specify-ordereddict-k-v-types-for-mypy-type-annotation
        result: collections.OrderedDict[Any, str] = collections.OrderedDict()
        for name, data in ImageStack.AXES_DATA.items():
            result[name] = self._data.shape[data.order]
        result['y'] = self._data.shape[-2]
        result['x'] = self._data.shape[-1]

        return result

    def _get_dimension_size(self, dimension: Indices):
        axis_data = ImageStack.AXES_DATA[dimension]
        if dimension in self._image_partition.dimensions or axis_data.required:
            return self._image_partition.get_dimension_shape(dimension)
        return 1

    @property
    def num_rounds(self):
        return self._get_dimension_size(Indices.ROUND)

    @property
    def num_chs(self):
        return self._get_dimension_size(Indices.CH)

    @property
    def num_zlayers(self):
        return self._get_dimension_size(Indices.Z)

    @property
    def tile_shape(self):
        return self._tile_shape

    def write(self, filepath: str, tile_opener=None) -> None:
        """write the image tensor to disk

        Parameters
        ----------
        filepath : str
            path + prefix for writing the image tensor
        tile_opener : TODO ttung: doc me.

        """
        if self._data_needs_writeback:
            for tile in self._image_partition.tiles():
                h = tile.indices[Indices.ROUND]
                c = tile.indices[Indices.CH]
                zlayer = tile.indices.get(Indices.Z, 0)
                tile.numpy_array, axes = self.get_slice(indices={Indices.ROUND: h, Indices.CH: c, Indices.Z: zlayer})
                assert len(axes) == 0
            self._data_needs_writeback = False

        seen_x_coords, seen_y_coords, seen_z_coords = set(), set(), set()
        for tile in self._image_partition.tiles():
            seen_x_coords.add(tile.coordinates[Coordinates.X])
            seen_y_coords.add(tile.coordinates[Coordinates.Y])
            z_coords = tile.coordinates.get(Coordinates.Z, None)
            if z_coords is not None:
                seen_z_coords.add(z_coords)

        sorted_x_coords = sorted(seen_x_coords)
        sorted_y_coords = sorted(seen_y_coords)
        sorted_z_coords = sorted(seen_z_coords)
        x_coords_to_idx = {coords: idx for idx, coords in enumerate(sorted_x_coords)}
        y_coords_to_idx = {coords: idx for idx, coords in enumerate(sorted_y_coords)}
        z_coords_to_idx = {coords: idx for idx, coords in enumerate(sorted_z_coords)}

        if tile_opener is None:
            def tile_opener(tileset_path, tile, ext):
                tile_basename = os.path.splitext(tileset_path)[0]
                xcoord = tile.coordinates[Coordinates.X]
                ycoord = tile.coordinates[Coordinates.Y]
                zcoord = tile.coordinates.get(Coordinates.Z, None)
                xcoord = tuple(xcoord) if isinstance(xcoord, list) else xcoord
                ycoord = tuple(ycoord) if isinstance(ycoord, list) else ycoord
                xval = x_coords_to_idx[xcoord]
                yval = y_coords_to_idx[ycoord]
                if zcoord is not None:
                    zval = z_coords_to_idx[zcoord]
                    zstr = "-Z{}".format(zval)
                else:
                    zstr = ""
                return open(
                    "{}-X{}-Y{}{}-H{}-C{}.{}".format(
                        tile_basename,
                        xval,
                        yval,
                        zstr,
                        tile.indices[Indices.ROUND],
                        tile.indices[Indices.CH],
                        ext,
                    ),
                    "wb")

        Writer.write_to_path(
            self._image_partition,
            filepath,
            pretty=True,
            tile_opener=tile_opener)

    def max_proj(self, *dims: Indices) -> np.ndarray:
        """return a max projection over one or more axis of the image tensor

        Parameters
        ----------
        dims : Indices
            one or more axes to project over

        Returns
        -------
        np.ndarray :
            max projection

        """
        max_projection = self._data.max([dim.value for dim in dims]).values
        return max_projection

    @staticmethod
    def _default_tile_extras_provider(round_: int, ch: int, z: int) -> Any:
        """
        Returns None for extras for any given round/ch/z.
        """
        return None

    @staticmethod
    def _default_tile_data_provider(round_: int, ch: int, z: int, height: int, width: int) -> np.ndarray:
        """
        Returns a tile of just ones for any given round/ch/z.
        """
        return np.ones((height, width))

    @classmethod
    def synthetic_stack(
            cls,
            num_round: int=4,
            num_ch: int=4,
            num_z: int=12,
            tile_height: int=50,
            tile_width: int=40,
            tile_data_provider: Callable[[int, int, int, int, int], np.ndarray]=None,
            tile_extras_provider: Callable[[int, int, int], Any]=None,
    ) -> "ImageStack":
        """generate a synthetic ImageStack

        Returns
        -------
        ImageStack :
            imagestack containing a tensor whose default shape is (2, 3, 4, 30, 20)
            and whose default values are all 1.

        """
        if tile_data_provider is None:
            tile_data_provider = cls._default_tile_data_provider
        if tile_extras_provider is None:
            tile_extras_provider = cls._default_tile_extras_provider

        img = TileSet(
            {Coordinates.X, Coordinates.Y, Indices.ROUND, Indices.CH, Indices.Z},
            {
                Indices.ROUND: num_round,
                Indices.CH: num_ch,
                Indices.Z: num_z,
            },
            default_tile_shape=(tile_height, tile_width),
        )
        for round_ in range(num_round):
            for ch in range(num_ch):
                for z in range(num_z):
                    tile = Tile(
                        {
                            Coordinates.X: (0.0, 0.001),
                            Coordinates.Y: (0.0, 0.001),
                            Coordinates.Z: (0.0, 0.001),
                        },
                        {
                            Indices.ROUND: round_,
                            Indices.CH: ch,
                            Indices.Z: z,
                        },
                        extras=tile_extras_provider(round_, ch, z),
                    )
                    tile.numpy_array = tile_data_provider(round_, ch, z, tile_height, tile_width)

                    img.add_tile(tile)

        stack = cls(img)
        return stack

    @classmethod
    def synthetic_spots(
            cls, intensities: IntensityTable, num_z: int, height: int, width: int,
            n_photons_background=1000, point_spread_function=(4, 2, 2),
            camera_detection_efficiency=0.25, background_electrons=1,
            graylevel: float=37000.0 / 2 ** 16, ad_conversion_bits=16,
    ) -> "ImageStack":
        """Generate a synthetic ImageStack from a set of Features stored in an IntensityTable

        Parameters
        ----------
        intensities : IntensityTable
            IntensityTable containing coordinates of fluorophores. Used to position and generate
            spots in the output ImageStack
        num_z : int
            Number of z-planes in the ImageStack
        height : int
            Height in pixels of the ImageStack
        width : int
            Width in pixels of the ImageStack
        n_photons_background : int
            Poisson rate for the number of background photons to add to each pixel of the image.
            Set this parameter to 0 to eliminate background.
            (default 1000)
        point_spread_function : Tuple[int]
            The width of the gaussian density wherein photons spread around their light source.
            Set to zero to eliminate this (default (4, 2, 2))
        camera_detection_efficiency : float
            The efficiency of the camera to detect light. Set to 1 to remove this filter (default
            0.25)
        background_electrons : int
            Poisson rate for the number of spurious electrons detected per pixel during image
            capture by the camera (default 1)
        graylevel : float
            The number of shades of gray displayable by the synthetic camera. Larger numbers will
            produce higher resolution images (default 37000 / 2 ** 16)
        ad_conversion_bits : int
            The number of bits used during analog to visual conversion (default 16)

        Returns
        -------

        """
        # check some params
        if not 0 < camera_detection_efficiency <= 1:
            raise ValueError(
                f'invalid camera_detection_efficiency value: {camera_detection_efficiency}. '
                f'Must be in the interval (0, 1].')

        def select_uint_dtype(array):
            """choose appropriate dtype based on values of an array"""
            max_val = np.max(array)
            for dtype in (np.uint8, np.uint16, np.uint32):
                if max_val <= np.iinfo(dtype).max:
                    return array.astype(dtype)
            raise ValueError('value exceeds dynamic range of largest skimage-supported type')

        # make sure requested dimensions are large enough to support intensity values
        indices = zip((Features.Z, Features.Y, Features.X), (num_z, height, width))
        for index, requested_size in indices:
            required_size = intensities.coords[index].values.max()
            if required_size > requested_size:
                raise ValueError(
                    f'locations of intensities contained in table exceed the size of requested '
                    f'dimension {index}. Required size {required_size} > {requested_size}.')

        # create an empty array of the correct size
        image = np.zeros(
            (
                intensities.sizes[Indices.ROUND.value],
                intensities.sizes[Indices.CH.value],
                num_z,
                height,
                width
            )
        )

        for ch, round_ in product(*(range(s) for s in intensities.shape[1:])):
            spots = intensities[:, ch, round_]

            # numpy deprecated casting a specific way of casting floats that is triggered in xarray
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', FutureWarning)
                values = spots.where(spots, drop=True)

            image[round_, ch, values.z, values.y, values.x] = values

        # add imaging noise
        image += np.random.poisson(n_photons_background, size=image.shape)

        # blur image over coordinates, but not over round_/channels (dim 0, 1)
        sigma = (0, 0) + point_spread_function
        image = gaussian_filter(image, sigma=sigma, mode='nearest')

        image *= camera_detection_efficiency

        image += np.random.normal(scale=background_electrons, size=image.shape)

        # mimic analog to digital conversion
        image = (image / graylevel).astype(int).clip(0, 2 ** ad_conversion_bits)

        # clip in case we've picked up some negative values
        image = np.clip(image, 0, a_max=None)

        # set correct dtype and convert to stack
        image = select_uint_dtype(image)
        return cls.from_numpy_array(image)

    def squeeze(self) -> np.ndarray:
        """return an array that is linear over categorical dimensions and z

        Returns
        -------
        np.ndarray :
            array of shape (num_rounds + num_channels + num_z_layers, x, y).

        """
        first_dim = self.num_rounds * self.num_chs * self.num_zlayers
        new_shape = (first_dim,) + self.tile_shape
        new_data = self.numpy_array.reshape(new_shape)

        return new_data

    def un_squeeze(self, stack):
        if type(stack) is list:
            stack = np.array(stack)

        new_shape = (self.num_rounds, self.num_chs, self.num_zlayers) + self.tile_shape
        res = stack.reshape(new_shape)
        return res

    @classmethod
    def from_numpy_array(cls, array: np.ndarray) -> "ImageStack":
        """Create an ImageStack from a 5d numpy array with shape (n_round, n_ch, n_z, y, x)

        Parameters
        ----------
        array : np.ndarray
            5-d tensor of shape (n_round, n_ch, n_z, y, x)

        Returns
        -------
        ImageStack :
            array data stored as an ImageStack

        """
        if len(array.shape) != 5:
            raise ValueError('a 5-d tensor with shape (n_round, n_ch, n_z, y, x) must be provided.')
        n_round, n_ch, n_z, height, width = array.shape
        empty = cls.synthetic_stack(
            num_round=n_round, num_ch=n_ch, num_z=n_z, tile_height=height, tile_width=width)

        # preserve original dtype
        empty._data = empty._data.astype(array.dtype)

        for h in np.arange(n_round):
            for c in np.arange(n_ch):
                for z in np.arange(n_z):
                    view = array[h, c, z]
                    empty.set_slice({Indices.ROUND: h, Indices.CH: c, Indices.Z: z}, view)

        return empty



import argparse
from functools import partial
from typing import Callable, Union, Tuple, Optional

import numpy as np
from skimage import img_as_uint

from starfish.errors import DataFormatWarning
from starfish.image import ImageStack
from starfish.pipeline.filter.gaussian_low_pass import GaussianLowPass
from starfish.types import Number
from ._base import FilterAlgorithmBase


class GaussianHighPass(FilterAlgorithmBase):

    def __init__(
            self, sigma: Union[Number, Tuple[Number]], is_volume: bool=False, verbose: bool=False,
            **kwargs
    ) -> None:
        """Gaussian high pass filter

        Parameters
        ----------
        sigma : Union[Number, Tuple[Number]]
            standard deviation of gaussian kernel
        is_volume : bool
            If True, 3d (z, y, x) volumes will be filtered, otherwise, filter 2d tiles
            independently.

        """
        if isinstance(sigma, tuple):
            message = ("if passing an anisotropic kernel, the dimensionality must match the data "
                       "shape ({shape}), not {passed_shape}")
            if is_volume and len(sigma) != 3:
                raise ValueError(message.format(shape=3, passed_shape=len(sigma)))
            if not is_volume and len(sigma) != 2:
                raise ValueError(message.format(shape=2, passed_shape=len(sigma)))

        self.sigma = sigma
        self.is_volume = is_volume
        self.verbose = verbose

    @classmethod
    def add_arguments(cls, group_parser: argparse.ArgumentParser) -> None:
        group_parser.add_argument(
            "--sigma", type=float, help="standard deviation of gaussian kernel")
        group_parser.add_argument(
            "--is-volume", action="store_true",
            help="indicates that the image stack should be filtered in 3d")

    @staticmethod
    def high_pass(image: np.ndarray, sigma: Union[Number, Tuple[Number]]) -> np.ndarray:
        """
        Applies a gaussian high pass filter to an image

        Parameters
        ----------
        image : numpy.ndarray[np.uint16]
            2-d or 3-d image data
        sigma : Union[Number, Tuple[Number]]
            Standard deviation of gaussian kernel

        Returns
        -------
        np.ndarray :
            filtered image of the same shape as the input image

        """
        if image.dtype != np.uint16:
            DataFormatWarning("gaussian filters currently only support uint16 images. Image data "
                              "will be converted.")
            image = img_as_uint(image)

        blurred: np.ndarray = GaussianLowPass.low_pass(image, sigma)

        over_flow_ind: np.ndarray[bool] = image < blurred
        filtered: np.ndarray = image - blurred
        filtered[over_flow_ind] = 0

        return filtered

    def run(
            self, stack: ImageStack, in_place: bool=True, verbose: bool=True
    ) -> Optional[ImageStack]:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on filtering progress (default = False)

        Returns
        -------
        Optional[ImageStack] :
            if in-place is False, return the results of filter as a new stack

        """
        high_pass: Callable = partial(self.high_pass, sigma=self.sigma)
        result = stack.apply(
            high_pass, is_volume=self.is_volume, verbose=self.verbose, in_place=in_place
        )
        if not in_place:
            return result
        return None



import warnings

import numpy as np
import pytest

from starfish.errors import DataFormatWarning
from starfish.image import ImageStack

NUM_ROUND = 2
NUM_CH = 2
NUM_Z = 2
HEIGHT = 10
WIDTH = 10


def create_tile_data_provider(dtype: np.number, corner_dtype: np.number):
    """
    Makes a stack that's all of the same type, except the round=0,ch=0,z=0 corner, which is a different type.  All the
    tiles are initialized with ones.

    Parameters
    ----------
    dtype : np.number
        The data type of all the tiles except the hyd=0,ch=0,z=0 corner.
    corner_dtype
        The data type of the tile in the hyd=0,ch=0,z=0 corner.

    Returns
    -------
    ImageStack :
        The image stack with the tiles initialized as described.
    """
    def tile_data_provider(round_: int, ch: int, z: int, height: int, width: int) -> np.ndarray:
        if round_ == 0 and ch == 0 and z == 0:
            return np.ones((height, width), dtype=corner_dtype)
        else:
            return np.ones((height, width), dtype=dtype)
    return tile_data_provider


def test_multiple_tiles_of_different_kind():
    with pytest.raises(TypeError):
        ImageStack.synthetic_stack(
            NUM_ROUND, NUM_CH, NUM_Z,
            HEIGHT, WIDTH,
            tile_data_provider=create_tile_data_provider(np.uint32, np.float32),
        )


def test_multiple_tiles_of_same_dtype():
    stack = ImageStack.synthetic_stack(
        NUM_ROUND, NUM_CH, NUM_Z,
        HEIGHT, WIDTH,
        tile_data_provider=create_tile_data_provider(np.uint32, np.uint32),
    )
    expected = np.ones(
        (NUM_ROUND,
         NUM_CH,
         NUM_Z,
         HEIGHT,
         WIDTH), dtype=np.uint32)
    assert np.array_equal(stack.numpy_array, expected)


def test_int_type_promotion():
    with warnings.catch_warnings(record=True) as w:
        stack = ImageStack.synthetic_stack(
            NUM_ROUND, NUM_CH, NUM_Z,
            HEIGHT, WIDTH,
            tile_data_provider=create_tile_data_provider(np.int32, np.int8),
        )
        assert len(w) == 1
        assert issubclass(w[0].category, DataFormatWarning)
    expected = np.ones(
        (NUM_ROUND,
         NUM_CH,
         NUM_Z,
         HEIGHT,
         WIDTH), dtype=np.int32)
    corner = np.empty(
        (HEIGHT,
         WIDTH), dtype=np.int32)
    corner.fill(16777216)
    expected[0, 0, 0] = corner
    assert np.array_equal(stack.numpy_array, expected)


def test_uint_type_promotion():
    with warnings.catch_warnings(record=True) as w:
        stack = ImageStack.synthetic_stack(
            NUM_ROUND, NUM_CH, NUM_Z,
            HEIGHT, WIDTH,
            tile_data_provider=create_tile_data_provider(np.uint32, np.uint8),
        )
        assert len(w) == 1
        assert issubclass(w[0].category, DataFormatWarning)
    expected = np.ones(
        (NUM_ROUND,
         NUM_CH,
         NUM_Z,
         HEIGHT,
         WIDTH), dtype=np.uint32)
    corner = np.empty(
        (HEIGHT,
         WIDTH), dtype=np.uint32)
    corner.fill(16777216)
    expected[0, 0, 0] = corner
    assert np.array_equal(stack.numpy_array, expected)


def test_float_type_promotion():
    with warnings.catch_warnings(record=True) as w:
        stack = ImageStack.synthetic_stack(
            NUM_ROUND, NUM_CH, NUM_Z,
            HEIGHT, WIDTH,
            tile_data_provider=create_tile_data_provider(np.float64, np.float32),
        )
        assert len(w) == 1
        assert issubclass(w[0].category, DataFormatWarning)
    expected = np.ones(
        (NUM_ROUND,
         NUM_CH,
         NUM_Z,
         HEIGHT,
         WIDTH), dtype=np.float64)
    assert np.array_equal(stack.numpy_array, expected)



from typing import Optional, Union

import numpy as np
import pandas as pd
import xarray as xr
from skimage.feature import blob_log

from starfish.constants import Features
from starfish.image import ImageStack
from starfish.intensity_table import IntensityTable
from starfish.pipeline.features.spot_attributes import SpotAttributes
from starfish.pipeline.features.spots.detector.detect import (
    measure_spot_intensity,
    detect_spots,
)
from starfish.types import Number
from starfish.util.argparse import FsExistsType
from ._base import SpotFinderAlgorithmBase


class GaussianSpotDetector(SpotFinderAlgorithmBase):

    def __init__(
            self, min_sigma: Number, max_sigma: Number, num_sigma: int, threshold: Number,
            overlap=0.5, measurement_type='max', is_volume: bool=True, **kwargs) \
            -> None:
        """Multi-dimensional gaussian spot detector

        This method is a wrapper for skimage.feature.blob_log

        Parameters
        ----------
        min_sigma : float
            The minimum standard deviation for Gaussian Kernel. Keep this low to
            detect smaller blobs.
        max_sigma : float
            The maximum standard deviation for Gaussian Kernel. Keep this high to
            detect larger blobs.
        num_sigma : int
            The number of intermediate values of standard deviations to consider
            between `min_sigma` and `max_sigma`.
        threshold : float
            The absolute lower bound for scale space maxima. Local maxima smaller
            than thresh are ignored. Reduce this to detect blobs with less
            intensities.
        overlap : float [0, 1]
            If two spots have more than this fraction of overlap, the spots are combined
            (default = 0.5)
        measurement_type : str ['max', 'mean']
            name of the function used to calculate the intensity for each identified spot area

        Notes
        -----
        # TODO ambrosejcarr: revisit after changing dtype assumptions of library to float in [0, 1]
        This spot detector is very sensitive to the threshold that is selected, and the threshold
        is defined as an absolute value -- therefore it must be adjusted depending on the datatype
        of the passed image.

        See Also
        --------
        http://scikit-image.org/docs/dev/api/skimage.feature.html#skimage.feature.blob_log

        """
        self.min_sigma = min_sigma
        self.max_sigma = max_sigma
        self.num_sigma = num_sigma
        self.threshold = threshold
        self.overlap = overlap
        self.is_volume = is_volume
        self.measurement_function = self._get_measurement_function(measurement_type)

    def image_to_spots(self, data_image: Union[np.ndarray, xr.DataArray]) -> SpotAttributes:
        """
        Find spots using a gaussian blob finding algorithm

        Parameters
        ----------
        data_image : Union[np.ndarray, xr.DataArray]
            ImageStack containing blobs to be detected

        Returns
        -------
        SpotAttributes :
            DataFrame of metadata containing the coordinates, intensity and radius of each spot

        """

        fitted_blobs_array: np.ndarray = blob_log(
            data_image,
            self.min_sigma,
            self.max_sigma,
            self.num_sigma,
            self.threshold,
            self.overlap
        )

        # create the SpotAttributes Table
        columns = [Features.Z, Features.Y, Features.X, Features.SPOT_RADIUS]
        fitted_blobs = pd.DataFrame(data=fitted_blobs_array, columns=columns)

        # convert standard deviation of gaussian kernel used to identify spot to radius of spot
        converted_radius = np.round(fitted_blobs[Features.SPOT_RADIUS] * np.sqrt(3))
        fitted_blobs[Features.SPOT_RADIUS] = converted_radius

        # convert the array to int so it can be used to index
        rounded_blobs: pd.DataFrame = fitted_blobs.astype(int)

        rounded_blobs['intensity'] = measure_spot_intensity(
            data_image, rounded_blobs, self.measurement_function)
        rounded_blobs['spot_id'] = np.arange(rounded_blobs.shape[0])

        return SpotAttributes(rounded_blobs)

    def find(
            self, data_stack: ImageStack,
            blobs_image: Optional[Union[np.ndarray, xr.DataArray]]=None,
            reference_image_from_max_projection: bool=False) \
            -> IntensityTable:
        """find spots in an ImageStack

        Parameters
        ----------
        data_stack : ImageStack
            stack containing spots to find
        blobs_image : Union[np.ndarray, xr.DataArray]
        reference_image_from_max_projection : bool
            if True, compute a reference image from the maximum projection of the channels and
            z-planes

        Returns
        -------
        IntensityTable :
            3d tensor containing the intensity of spots across channels and imaging rounds

        """

        intensity_table = detect_spots(
            data_stack=data_stack,
            spot_finding_method=self.image_to_spots,
            reference_image=blobs_image,
            reference_image_from_max_projection=reference_image_from_max_projection,
            measurement_function=self.measurement_function
        )

        return intensity_table

    @classmethod
    def add_arguments(cls, group_parser):
        group_parser.add_argument("--blobs-stack", type=FsExistsType(), required=True)
        group_parser.add_argument(
            "--min-sigma", default=4, type=int, help="Minimum spot size (in standard deviation)")
        group_parser.add_argument(
            "--max-sigma", default=6, type=int, help="Maximum spot size (in standard deviation)")
        group_parser.add_argument("--num-sigma", default=20, type=int, help="Number of sigmas to try")
        group_parser.add_argument("--threshold", default=.01, type=float, help="Dots threshold")
        group_parser.add_argument(
            "--overlap", default=0.5, type=float, help="dots with overlap of greater than this fraction are combined")
        group_parser.add_argument(
            "--show", default=False, action='store_true', help="display results visually")



import argparse
import os

from starfish.image import ImageStack
from starfish.pipeline.pipelinecomponent import PipelineComponent
from starfish.util.argparse import FsExistsType
from starfish.constants import Indices
from . import _base
from . import gaussian
from . import local_max_peak_finder


class SpotFinder(PipelineComponent):

    spot_finder_group: argparse.ArgumentParser

    @classmethod
    def implementing_algorithms(cls):
        return _base.SpotFinderAlgorithmBase.__subclasses__()

    @classmethod
    def add_to_parser(cls, subparsers):
        """Adds the spot finder component to the CLI argument parser."""
        spot_finder_group = subparsers.add_parser("detect_spots")
        spot_finder_group.add_argument("-i", "--input", type=FsExistsType(), required=True)
        spot_finder_group.add_argument("-o", "--output", required=True)
        spot_finder_group.add_argument(
            "--blobs-stack", default=None, required=False,
            help=(
                'ImageStack that contains the blobs. Will be max-projected across imaging round '
                'and channel to produce the blobs_image'
            )
        )
        spot_finder_group.add_argument(
            "--reference-image-from-max-projection", default=False, action='store_true',
            help=(
                'Construct a reference image by max projecting imaging rounds and channels. Spots '
                'are found in this image and then measured across all images in the input stack.'
            )
        )
        spot_finder_group.set_defaults(starfish_command=SpotFinder._cli)
        spot_finder_subparsers = spot_finder_group.add_subparsers(
            dest="spot_finder_algorithm_class")

        for algorithm_cls in cls.algorithm_to_class_map().values():
            group_parser = spot_finder_subparsers.add_parser(algorithm_cls.get_algorithm_name())
            group_parser.set_defaults(spot_finder_algorithm_class=algorithm_cls)
            algorithm_cls.add_arguments(group_parser)

        cls.spot_finder_group = spot_finder_group

    @classmethod
    def _cli(cls, args, print_help=False):
        """Runs the spot finder component based on parsed arguments."""

        if args.spot_finder_algorithm_class is None or print_help:
            cls.spot_finder_group.print_help()
            cls.spot_finder_group.exit(status=2)

        print('Detecting Spots ...')
        image_stack = ImageStack.from_path_or_url(args.input)
        if args.blobs_stack is not None:
            blobs_stack = ImageStack.from_path_or_url(args.blobs_stack)  # type: ignore
            blobs_image = blobs_stack.max_proj(Indices.ROUND, Indices.CH)
        else:
            blobs_image = None
        instance = args.spot_finder_algorithm_class(**vars(args))
        intensities = instance.find(
            image_stack,
            blobs_image=blobs_image,
            reference_image_from_max_projection=args.reference_image_from_max_projection
        )
        intensities.save(os.path.join(args.output, 'spots.nc'))



from ._base import DecoderAlgorithmBase


class IssDecoder(DecoderAlgorithmBase):
    def __init__(self, **kwargs):
        pass

    @classmethod
    def add_arguments(cls, group_parser):
        pass

    def decode(self, intensities, codebook):
        return codebook.decode_per_round_max(intensities)



from functools import partial
from typing import Optional

import numpy as np

from starfish.image import ImageStack
from ._base import FilterAlgorithmBase


class Clip(FilterAlgorithmBase):

    def __init__(self, p_min: int=0, p_max: int=100, is_volume: bool=False, **kwargs) -> None:
        """Image clipping filter

        Parameters
        ----------
        p_min : int
            values below this percentile are set to p_min (default 0)
        p_max : int
            values above this percentile are set to p_max (default 100)
        is_volume : bool
            If True, 3d (z, y, x) volumes will be filtered. By default, filter 2-d (y, x) tiles

        kwargs
        """
        self.p_min = p_min
        self.p_max = p_max
        self.is_volume = is_volume

    @classmethod
    def add_arguments(cls, group_parser) -> None:
        group_parser.add_argument(
            "--p-min", default=0, type=int, help="clip intensities below this percentile")
        group_parser.add_argument(
            "--p-max", default=100, type=int, help="clip intensities above this percentile")

    @staticmethod
    def clip(image: np.ndarray, p_min: int, p_max: int) -> np.ndarray:
        """Clip values of img below and above percentiles p_min and p_max

        Parameters
        ----------
        image : np.ndarray
            image to be clipped
        p_min : int
          values below this percentile are set to the value of this percentile
        p_max : int
          values above this percentile are set to the value of this percentile

        Notes
        -----
        - Wrapper for np.clip
        - No shifting or transformation to adjust dynamic range is done after clipping

        Returns
        -------
        np.ndarray :
          Numpy array of same shape as img

        """
        v_min, v_max = np.percentile(image, [p_min, p_max])

        # asking for a float percentile clipping value from an integer image will
        # convert to float, so store the dtype so it can be restored
        dtype = image.dtype
        image = image.clip(min=v_min, max=v_max)
        return image.astype(dtype)

    def run(
            self, stack: ImageStack, in_place: bool=True, verbose: bool=False
    ) -> Optional[ImageStack]:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            If True, report on the percentage completed (default = False) during processing

        Returns
        -------
        Optional[ImageStack] :
            if in-place is False, return the results of filter as a new stack

        """
        clip = partial(self.clip, p_min=self.p_min, p_max=self.p_max)
        result = stack.apply(clip, is_volume=self.is_volume, verbose=verbose, in_place=in_place)
        if not in_place:
            return result
        return None



from typing import List, NamedTuple, Tuple

import numpy as np
import pandas as pd
from skimage.measure import regionprops, label

from starfish.constants import Indices, Features
from starfish.intensity_table import IntensityTable
from starfish.munge import dataframe_to_multiindex
from starfish.types import Number


class ConnectedComponentDecodingResult(NamedTuple):
    region_properties: List
    label_image: np.ndarray
    decoded_image: np.ndarray


def combine_adjacent_features(
        intensities: IntensityTable, min_area: Number, max_area: Number, connectivity: int=2,
) -> Tuple[IntensityTable, ConnectedComponentDecodingResult]:
    """

    Parameters
    ----------
    intensities : IntensityTable
        pixel-valued IntensityTable
    min_area : Number
        discard spots whose total pixel area is less than this value
    max_area : Number
        discard spots whose total pixel area is greater than this value
    connectivity : int
        the connectivity allowed when constructing connected components. Intuitively it can be
        thought of as the number of square steps allowed between pixels that are considered
        connected. 1 enforces adjacent pixels. 2 (default) allows diagonal pixels in the same
        plane. 3 allows diagonal pixels in 3d.

    Returns
    -------
    IntensityTable :
        spot intensities
    ConnectedComponentDecodingResult :
        region properties for each spot (see skimage.measure.regionprops)
        label image wherein each connected component (spot) is coded with a different integer
        decoded image wherein each target is coded as a different integer. Intended for
        visualization

    """
    # None needs to map to zero, non-none needs to map to something else.
    int_to_target = dict(
        zip(range(1, np.iinfo(np.int).max),
            set(intensities[Features.AXIS][Features.TARGET].values) - {'None'}))
    int_to_target[0] = 'None'
    target_to_int = {v: k for (k, v) in int_to_target.items()}

    # map targets to ints
    target_list = [
        target_to_int[g] for g in intensities.coords[Features.TARGET].values]
    target_array = np.array(target_list)

    # reverses the linearization that was used to construct the IntensityTable from the ImageStack
    decoded_image: np.ndarray = target_array.reshape(intensities.attrs[IntensityTable.IMAGE_SHAPE])

    # label each pixel according to its component
    label_image: np.ndarray = label(decoded_image, connectivity=connectivity)

    # calculate the mean value grouped by (ch, round) across all pixels of a connected component
    SPOT_ID = 'spot_id'
    pixel_labels = label_image.reshape(-1)
    intensities[SPOT_ID] = (Features.AXIS, pixel_labels)
    mean_pixel_traces = intensities.groupby(SPOT_ID).mean(Features.AXIS)
    mean_distances = intensities[Features.DISTANCE].groupby(SPOT_ID).mean(Features.AXIS)
    mean_pixel_traces[Features.DISTANCE] = (
        SPOT_ID,
        np.ravel(mean_distances)
    )

    # "Zero" spot id is background in the label_image. Remove it.
    # TODO ambrosejcarr: can we replace zero with nan above to get around this?
    mean_pixel_traces = mean_pixel_traces.loc[mean_pixel_traces[SPOT_ID] > 0]
    props: List = regionprops(np.squeeze(label_image))

    # calculate spots and drop ones that fail the area threshold
    spots = []
    passes_filter = np.ones_like(props, dtype=np.bool)  # default is passing (ones)
    for i, spot_property in enumerate(props):
        # TODO ambrosejcarr: add tests for min/max area when fixing masking
        if min_area <= spot_property.area < max_area:

            # because of the above skimage issue, we need to support both 2d and 3d properties
            if len(spot_property.centroid) == 3:
                spot_attrs = {
                    'z': int(spot_property.centroid[0]),
                    'y': int(spot_property.centroid[1]),
                    'x': int(spot_property.centroid[2])
                }
            else:  # data is 2d
                spot_attrs = {
                    'z': 0,
                    'y': int(spot_property.centroid[0]),
                    'x': int(spot_property.centroid[1])
                }

            # we're back to 3d or fake-3d here
            target_index = decoded_image[spot_attrs['z'], spot_attrs['y'], spot_attrs['x']]
            spot_attrs[Features.TARGET] = int_to_target[target_index]
            spot_attrs[Features.SPOT_RADIUS] = spot_property.equivalent_diameter / 2
            spots.append(spot_attrs)
        else:
            passes_filter[i] = 0  # did not pass radius filter

    # filter intensities
    mean_pixel_traces = mean_pixel_traces.loc[passes_filter]

    # now I need to make an IntensityTable from this thing.
    spots_df = pd.DataFrame(spots)
    spots_df[Features.DISTANCE] = \
        mean_pixel_traces[Features.DISTANCE]

    # create new indexes for the output IntensityTable
    spots_index = dataframe_to_multiindex(spots_df)
    channel_index = mean_pixel_traces.indexes[Indices.CH]
    round_index = mean_pixel_traces.indexes[Indices.ROUND]

    # create the output IntensityTable
    dims = (Features.AXIS, Indices.CH.value, Indices.ROUND.value)
    attrs = {IntensityTable.IMAGE_SHAPE: intensities.attrs[IntensityTable.IMAGE_SHAPE]}
    intensity_table = IntensityTable(
        data=mean_pixel_traces, coords=(spots_index, channel_index, round_index), dims=dims,
        attrs=attrs
    )

    ccdr = ConnectedComponentDecodingResult(props, label_image, decoded_image)

    return intensity_table, ccdr



import argparse
from functools import partial
from typing import Callable, Optional, Union, Tuple

import numpy as np
from scipy.ndimage.filters import uniform_filter
from skimage import img_as_uint

from starfish.errors import DataFormatWarning
from starfish.image import ImageStack
from ._base import FilterAlgorithmBase
from starfish.types import Number


class MeanHighPass(FilterAlgorithmBase):

    def __init__(
            self, size: Union[Number, Tuple[Number]], is_volume: bool=False, **kwargs) -> None:
        """Mean high pass filter.

        The mean high pass filter reduces low spatial frequency features by subtracting a
        mean filtered image from the original image. The mean filter smooths an image by replacing
        each pixel's value with an average of the pixel values of the surrounding neighborhood.

        The mean filter is also known as a uniform or box filter.

        This is a pass through for the scipy.ndimage.filters.uniform_filter:
        https://docs.scipy.org/doc/scipy-0.19.0/reference/generated/scipy.ndimage.uniform_filter.html

        Parameters
        ----------
        size : Union[Number, Tuple[Number]]
            width of the kernel
        is_volume : bool
            If True, 3d (z, y, x) volumes will be filtered, otherwise, filter 2d tiles
            independently.

        """

        if isinstance(size, tuple):
            message = ("if passing an anisotropic kernel, the dimensionality must match the data "
                       "shape ({shape}), not {passed_shape}")
            if is_volume and len(size) != 3:
                raise ValueError(message.format(shape=3, passed_shape=len(size)))
            if not is_volume and len(size) != 2:
                raise ValueError(message.format(shape=2, passed_shape=len(size)))

        self.size = size
        self.is_volume = is_volume

    @classmethod
    def add_arguments(cls, group_parser: argparse.ArgumentParser) -> None:
        group_parser.add_argument(
            "--size", type=float, help="width of the kernel")
        group_parser.add_argument(
            "--is-volume", action="store_true", help="indicates that the image stack should be filtered in 3d")

    @staticmethod
    def high_pass(image: np.ndarray, size: Number) -> np.ndarray:
        """
        Applies a gaussian high pass filter to an image

        Parameters
        ----------
        image : numpy.ndarray[np.uint16]
            2-d or 3-d image data
        size : Number
            width of the kernel

        Returns
        -------
        np.ndarray :
            Filtered image, same shape as input

        """
        if image.dtype != np.uint16:
            DataFormatWarning(
                "Mean filters currently only support uint16 images. Image data will be converted.")
            image = img_as_uint(image)

        blurred: np.ndarray = uniform_filter(image, size)

        over_flow_ind: np.ndarray[bool] = image < blurred
        filtered: np.ndarray = image - blurred
        filtered[over_flow_ind] = 0

        return filtered

    def run(
            self, stack: ImageStack, in_place: bool=True, verbose: bool=False
    ) -> Optional[ImageStack]:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on filtering progress (default = False)

        Returns
        -------
        Optional[ImageStack] :
            if in_place is False, return the results of filter as a new stack

        """
        high_pass: Callable = partial(self.high_pass, size=self.size)
        result = stack.apply(
            high_pass, is_volume=self.is_volume, verbose=verbose, in_place=in_place)
        if not in_place:
            return result
        return None



import json
from typing import Optional, Mapping

import numpy as np
from semantic_version import Version
from slicedimage.io import resolve_path_or_url

from .image import ImageStack


class Stack:
    MIN_SUPPORTED_VERSION = Version("1.0.0")
    MAX_SUPPORTED_VERSION = Version("1.0.0")

    def __init__(self):
        # data organization
        self.org = None
        self.image = None

        # auxiliary images
        self.auxiliary_images = dict()

        # readers and writers
        self.write_fn = np.save  # asserted for now

        # backend & baseurl
        self.backend = None
        self.baseurl = None

    @classmethod
    def from_data(cls, image_stack: ImageStack, aux_dict: Optional[Mapping[str, ImageStack]]=None) -> "Stack":
        """create a Stack from an already-loaded ImageStack

        Parameters
        ----------
        image_stack : ImageStack
            in-memory ImageStack
        aux_dict : Optional[Mapping[str, ImageStack]]
            a dictionary of ImageStacks, default None

        Returns
        -------
        Stack :
            a Stack object

        """
        stack = cls()
        stack.image = image_stack
        stack.auxiliary_images = aux_dict if aux_dict is not None else dict()
        return stack

    def read(self, in_json_path_or_url):
        self.backend, name, self.baseurl = resolve_path_or_url(in_json_path_or_url)
        with self.backend.read_file_handle(name) as fh:
            self.org = json.load(fh)

        self.verify_version(self.org['version'])
        self.image = ImageStack.from_url(self.org['hybridization_images'], self.baseurl)
        for aux_key, aux_data in self.org['auxiliary_images'].items():
            self.auxiliary_images[aux_key] = ImageStack.from_url(aux_data, self.baseurl)

    @classmethod
    def verify_version(cls, semantic_version_str: str):
        version = Version(semantic_version_str)
        if not (Stack.MIN_SUPPORTED_VERSION <= version <= Stack.MAX_SUPPORTED_VERSION):
            raise ValueError(
                f"version {version} not supported.  This version of the starfish library only "
                f"supports formats from {Stack.MIN_SUPPORTED_VERSION} to "
                f"{Stack.MAX_SUPPORTED_VERSION}")

    @classmethod
    def from_experiment_json(cls, json_url: str) -> "Stack":
        """Construct a `Stack` from an experiment.json file format specifier

        Parameters
        ----------
        json_url : str
            file path or web link to an experiment.json file

        Returns
        -------
        Stack :
            Stack object serving the requested image data

        """
        stack: Stack = cls()
        stack.read(json_url)
        return stack



import argparse

from starfish.image import ImageStack
from starfish.pipeline.pipelinecomponent import PipelineComponent
from starfish.util.argparse import FsExistsType
from . import fourier_shift
from ._base import RegistrationAlgorithmBase


class Registration(PipelineComponent):

    register_group: argparse.ArgumentParser

    @classmethod
    def implementing_algorithms(cls):
        return RegistrationAlgorithmBase.__subclasses__()

    @classmethod
    def add_to_parser(cls, subparsers):
        """Adds the registration component to the CLI argument parser."""
        register_group = subparsers.add_parser("registration")
        register_group.add_argument("-i", "--input", type=FsExistsType(), required=True)
        register_group.add_argument("-o", "--output", required=True)
        register_group.set_defaults(starfish_command=Registration._cli)
        registration_subparsers = register_group.add_subparsers(dest="registration_algorithm_class")

        for algorithm_cls in cls.algorithm_to_class_map().values():
            group_parser = registration_subparsers.add_parser(algorithm_cls.get_algorithm_name())
            group_parser.set_defaults(registration_algorithm_class=algorithm_cls)
            algorithm_cls.add_arguments(group_parser)

        cls.register_group = register_group

    @classmethod
    def _cli(cls, args, print_help=False):
        """Runs the registration component based on parsed arguments."""
        if args.registration_algorithm_class is None or print_help:
            cls.register_group.print_help()
            cls.register_group.exit(status=2)

        print('Registering ...')
        stack = ImageStack.from_path_or_url(args.input)
        instance = args.registration_algorithm_class(**vars(args))
        instance.run(stack)

        stack.write(args.output)



import warnings
import tempfile
import json
import os
import numpy as np
import pytest
from unittest.mock import patch, MagicMock
import xarray as xr

from starfish.constants import Indices, Features
from starfish.codebook import Codebook

# don't inspect pytest fixtures in pycharm
# noinspection PyUnresolvedReferences
from starfish.test.dataset_fixtures import (
    simple_codebook_json, simple_codebook_array, euclidean_decoded_intensities,
    per_channel_max_decoded_intensities, loaded_codebook, small_intensity_table)


def test_loading_codebook_from_json_local_file(simple_codebook_json):
    cb = Codebook.from_json(simple_codebook_json, n_ch=2, n_round=2)
    assert isinstance(cb, Codebook)


@patch('starfish.codebook.urllib.request.urlopen')
def test_loading_codebook_from_json_https_file(mock_urlopen):

    # codebook data to pass to the mock
    _return_value = json.dumps(
        [
            {
                Features.CODEWORD: [
                    {Indices.ROUND.value: 0, Indices.CH.value: 0, Features.CODE_VALUE: 1},
                    {Indices.ROUND.value: 1, Indices.CH.value: 1, Features.CODE_VALUE: 1}
                ],
                Features.TARGET: "SCUBE2"
            }
        ]
    ).encode()

    # mock urlopen.read() to return data corresponding to a codebook
    a = MagicMock()
    a.read.side_effect = [_return_value]
    a.__enter__.return_value = a
    mock_urlopen.return_value = a

    # test that the function loads the codebook from the link when called
    cb = Codebook.from_json('https://www.alink.com/file.json', n_ch=2, n_round=2)
    assert isinstance(cb, Codebook)
    assert mock_urlopen.call_count == 1


def test_loading_codebook_from_list(simple_codebook_array):
    cb = Codebook.from_code_array(simple_codebook_array, n_ch=2, n_round=2)
    assert isinstance(cb, Codebook)


def test_loading_codebook_without_specifying_ch_round_guesses_correct_values(simple_codebook_array):
    cb = Codebook.from_code_array(simple_codebook_array)
    assert cb.shape == (3, 2, 2)


def test_loading_codebook_with_too_few_dims_raises_value_error(simple_codebook_json):
    with pytest.raises(ValueError):
        Codebook.from_json(simple_codebook_json, n_ch=1, n_round=2)

    with pytest.raises(ValueError):
        Codebook.from_json(simple_codebook_json, n_ch=2, n_round=1)


def test_euclidean_decode_yields_correct_output(euclidean_decoded_intensities):
    expected_target_annotation = np.array(["ACTB", "SCUBE2", "BRCA", "None", "SCUBE2"])
    observed_target_annotation = euclidean_decoded_intensities[
        Features.TARGET].values
    expected_distances = np.array([0, 0, 0, 0.76536686, 0])
    observed_distances = euclidean_decoded_intensities[
        Features.DISTANCE].values
    assert np.array_equal(expected_target_annotation, observed_target_annotation)
    assert np.allclose(expected_distances, observed_distances)


def test_indexing_on_set_targets(euclidean_decoded_intensities):
    # note that this kind of indexing produces an xarray-internal FutureWarning about float
    # conversion that we can safely ignore here.
    is_actin = euclidean_decoded_intensities[Features.TARGET] == 'ACTB'
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', FutureWarning)

        # select only the intensities that are actin, drop the rest
        result = euclidean_decoded_intensities.where(is_actin, drop=True)

    assert result.shape == (1, 2, 2)


def test_per_channel_max_decode_yields_expected_results(per_channel_max_decoded_intensities):
    expected_target_annotation = np.array(["ACTB", "SCUBE2", "BRCA", "None", "SCUBE2"])
    observed_target_annotation = per_channel_max_decoded_intensities[
        Features.TARGET].values
    assert np.array_equal(expected_target_annotation, observed_target_annotation)


def test_synthetic_one_hot_codes_produce_one_channel_per_round():
    cb = Codebook.synthetic_one_hot_codebook(n_round=6, n_channel=4, n_codes=100)
    # sum over channels: only one should be "on"
    assert np.all(cb.sum(Indices.CH.value) == 1)


def test_codebook_save(loaded_codebook):
    directory = tempfile.mkdtemp()
    filename = os.path.join(directory, 'codebook.json')
    loaded_codebook.to_json(filename)
    reloaded = Codebook.from_json(filename, n_round=2, n_ch=2)

    assert np.array_equal(loaded_codebook, reloaded)
    assert np.array_equal(loaded_codebook[Indices.CH.value], reloaded[Indices.CH.value])
    assert np.array_equal(loaded_codebook[Indices.ROUND.value], reloaded[Indices.ROUND.value])
    assert np.array_equal(loaded_codebook[Features.TARGET].values,
                          reloaded[Features.TARGET].values)


@pytest.mark.parametrize('n_ch, n_round', ((2, 2), (5, 4)))
def test_loading_codebook_with_unused_channels_and_rounds(simple_codebook_json, n_ch, n_round):
    cb = Codebook.from_json(simple_codebook_json, n_ch=n_ch, n_round=n_round)
    assert cb.shape == (3, n_ch, n_round)


@pytest.mark.parametrize('n_ch, n_round', ((2, 2), (5, 4)))
def test_code_length(n_ch, n_round):
    target_names = np.arange(10)
    cb = Codebook._empty_codebook(target_names, n_ch, n_round)
    assert cb.code_length == n_ch * n_round


def test_unit_normalize():
    simple_data = np.array(
        [[[0, 1],
          [2, 1]],
         [[0, 0],
          [0, 0]],
         [[9, 0],
          [0, 1]]]
    )
    simple_codebook = Codebook(
        data=simple_data,
        dims=([Features.AXIS, Indices.CH.value, Indices.ROUND.value]),
        coords=(list('xyz'), list('vw'), list('tu'))
    )

    results, _ = Codebook._normalize_features(simple_codebook, norm_order=1)
    assert np.array_equal(results.sum([Indices.CH, Indices.ROUND]), np.ones(3))



from typing import Tuple

import numpy as np
import pandas as pd

from starfish.constants import Indices, Features
from starfish.intensity_table import IntensityTable
from starfish.codebook import Codebook
from starfish.image import ImageStack
from starfish.pipeline.features.pixels.combine_adjacent_features import combine_adjacent_features
# don't inspect pytest fixtures in pycharm
# noinspection PyUnresolvedReferences
from starfish.test.dataset_fixtures import (
    loaded_codebook, simple_codebook_json, simple_codebook_array, single_synthetic_spot)


def test_empty_intensity_table():
    x = [1, 2]
    y = [2, 3]
    z = [1, 1]
    r = [1, 1]
    spot_attributes = pd.MultiIndex.from_arrays(
        [x, y, z, r],
        names=(Features.X, Features.Y, Features.Z, Features.SPOT_RADIUS))
    image_shape = (2, 4, 3)
    empty = IntensityTable.empty_intensity_table(spot_attributes, 2, 2, image_shape)
    assert empty.shape == (2, 2, 2)
    assert np.sum(empty.values) == 0


def test_synthetic_intensities_generates_correct_number_of_features(loaded_codebook):
    n_spots = 2
    intensities = IntensityTable.synthetic_intensities(loaded_codebook, n_spots=n_spots)
    assert isinstance(intensities, IntensityTable)

    # shape should have n_spots and channels and imaging rounds equal to the codebook's shape
    assert intensities.shape == (n_spots, *loaded_codebook.shape[1:])


def test_synthetic_intensities_have_correct_number_of_on_features(loaded_codebook):
    n_spots = 2
    intensities = IntensityTable.synthetic_intensities(loaded_codebook, n_spots=n_spots)
    on_features = np.sum(intensities.values != 0)
    # this asserts that the number of features "on" in intensities should be equal to the
    # number of "on" features in the codewords, times the total number of spots in intensities.
    num_on = loaded_codebook.sum((Indices.CH.value, Indices.ROUND.value)).values.mean() * n_spots
    assert on_features == num_on


def feature_data() -> Tuple[Codebook, ImageStack]:
    # This codebook has two codes: on/off and on/on
    # This array has 3 spots: one on/off, one on/on, and one off/on
    # They exist in the first and second z-slice, but not the third.
    code_array = [
        {
            # on/off
            Features.CODEWORD: [
                {Indices.ROUND.value: 0, Indices.CH.value: 0, Features.CODE_VALUE: 1},
            ],
            Features.TARGET: "gene_1"
        },
        {
            # on/on
            Features.CODEWORD: [
                {Indices.ROUND.value: 0, Indices.CH.value: 0, Features.CODE_VALUE: 1},
                {Indices.ROUND.value: 1, Indices.CH.value: 0, Features.CODE_VALUE: 1},
            ],
            Features.TARGET: "gene_2"
        }
    ]
    codebook = Codebook.from_code_array(code_array)

    data = np.array(
        [[[[1, 1, 0, 1],  # round 0
           [1, 1, 0, 1],
           [0, 0, 0, 0]],

          [[1, 1, 0, 1],
           [1, 1, 0, 1],
           [0, 0, 0, 0]],

          [[0, 0, 0, 1],
           [0, 0, 0, 1],
           [0, 0, 0, 0]]],

         [[[1, 1, 0, 0],  # round 1
           [1, 1, 0, 0],
           [0, 0, 0, 1]],

          [[1, 1, 0, 0],
           [1, 1, 0, 0],
           [0, 0, 0, 1]],

          [[0, 0, 0, 0],
           [0, 0, 0, 0],
           [0, 0, 0, 1]]]]
    )
    data = data.reshape(2, 1, 3, 3, 4)
    image = ImageStack.from_numpy_array(data)
    return codebook, image


def test_combine_adjacent_features():
    codebook, image = feature_data()
    new_intensities = IntensityTable.from_image_stack(image)

    # TODO ambrosejcarr: this decoder is dropping pixels that don't decode. Desirable?
    new_intensities = codebook.metric_decode(
        new_intensities,
        max_distance=0.5,
        min_intensity=0.5,
        norm_order=2
    )

    # this is "working", with the caveat that the z-coord is a bit weird and potentially wrong.
    combined_intensities, _ = combine_adjacent_features(new_intensities, min_area=0, max_area=10)

    assert np.array_equal(combined_intensities.shape, (2, 1, 2))
    assert np.array_equal(
        combined_intensities[Features.AXIS][Features.TARGET],
        ['gene_2', 'gene_1']
    )



import argparse
import json
from typing import List, Type

from starfish.intensity_table import IntensityTable
from starfish.pipeline.pipelinecomponent import PipelineComponent
from starfish.util.argparse import FsExistsType
from . import point_in_poly
from ._base import TargetAssignmentAlgorithm


class TargetAssignment(PipelineComponent):

    target_assignment_group: argparse.ArgumentParser

    @classmethod
    def implementing_algorithms(cls) -> List[Type[TargetAssignmentAlgorithm]]:
        return TargetAssignmentAlgorithm.__subclasses__()

    @classmethod
    def add_to_parser(cls, subparsers) -> None:
        """Adds the target_assignment component to the CLI argument parser."""
        target_assignment_group = subparsers.add_parser("target_assignment")
        target_assignment_group.add_argument(
            "--coordinates-geojson", type=FsExistsType(), required=True)
        target_assignment_group.add_argument("--intensities", type=FsExistsType(), required=True)
        target_assignment_group.add_argument("-o", "--output", required=True)
        target_assignment_group.set_defaults(starfish_command=TargetAssignment._cli)
        target_assignment_group = target_assignment_group.add_subparsers(
            dest="target_assignment_algorithm_class")

        for algorithm_cls in cls.algorithm_to_class_map().values():
            group_parser = target_assignment_group.add_parser(algorithm_cls.get_algorithm_name())
            group_parser.set_defaults(target_assignment_algorithm_class=algorithm_cls)
            algorithm_cls.add_arguments(group_parser)

        cls.target_assignment_group = target_assignment_group

    @classmethod
    def _cli(cls, args, print_help=False) -> None:
        """Runs the target_assignment component based on parsed arguments."""
        from starfish import munge

        if args.target_assignment_algorithm_class is None or print_help:
            cls.target_assignment_group.print_help()
            cls.target_assignment_group.exit(status=2)

        with open(args.coordinates_geojson, "r") as fh:
            coordinates = json.load(fh)
        regions = munge.geojson_to_region(coordinates)

        print('Assigning targets to cells...')
        intensity_table = IntensityTable.load(args.intensities)

        instance = args.target_assignment_algorithm_class(**vars(args))

        result = instance.run(intensity_table, regions)

        print("Writing | cell_id | spot_id to: {}".format(args.output))
        result.to_json(args.output, orient="records")



import numpy as np
import pandas as pd
import regional
from skimage.measure import points_in_poly
from tqdm import tqdm

from starfish.constants import Features
from starfish.intensity_table import IntensityTable
from ._base import TargetAssignmentAlgorithm


class PointInPoly2D(TargetAssignmentAlgorithm):

    def __init__(self, **kwargs) -> None:
        """
        PointInPoly accepts no parameters, but all pipeline components must accept arbitrary kwargs
        """

    @classmethod
    def add_arguments(cls, parser) -> None:
        pass

    @staticmethod
    def _assign(
            cells_region: regional.many, spots: pd.DataFrame, use_hull: bool=True, verbose:
            bool=False
    ) -> pd.DataFrame:

        results = pd.DataFrame({'spot_id': range(0, spots.shape[0])})
        results['cell_id'] = None

        if verbose:
            cell_iterator = tqdm(range(cells_region.count))
        else:
            cell_iterator = range(cells_region.count)

        for cell_id in cell_iterator:
            if use_hull:
                vertices = cells_region[cell_id].hull
            else:
                vertices = cells_region[cell_id].coordinates
            vertices = np.array(vertices)
            in_poly = points_in_poly(spots, vertices)
            results.loc[results.spot_id[in_poly], 'cell_id'] = cell_id

        return results

    def run(
            self, intensity_table: IntensityTable, regions: regional.many, verbose: bool=False
    ) -> pd.DataFrame:
        """Assign spots with target assignments to cells

        Parameters
        ----------
        intensity_table : IntensityTable
        regions : regional.many
            # TODO dganguli can you add an explanation? I'll fix during PR.
        verbose : bool
            If True, report on the progress of gene assignment (default False)

        Returns
        -------
        pd.DataFrame :
            DataFrame mapping of spot ids to cell ids
            # TODO should this be emitted as an extra column of IntensityTable, instead?

        """
        # TODO must support filtering on the passes filter column
        # TODO does this support 3d assignment?
        x = intensity_table.coords[Features.AXIS][Features.X].values
        y = intensity_table.coords[Features.AXIS][Features.Y].values
        points = pd.DataFrame(dict(x=x, y=y))
        return self._assign(regions, points, use_hull=True, verbose=verbose)



from typing import Tuple

from starfish.codebook import Codebook
from starfish.image import ImageStack
from starfish.intensity_table import IntensityTable
from starfish.pipeline.features.pixels.combine_adjacent_features import \
    ConnectedComponentDecodingResult, combine_adjacent_features
from ._base import PixelFinderAlgorithmBase


class PixelSpotDetector(PixelFinderAlgorithmBase):
    def __init__(
            self, codebook: Codebook, metric: str, distance_threshold: float,
            magnitude_threshold: int, min_area: int, max_area: int, norm_order: int=2,
            crop_x: int=0, crop_y: int=0, crop_z: int=0, **kwargs) -> None:
        """Decode an image by first coding each pixel, then combining the results into spots

        Parameters
        ----------
        codebook : Codebook
            Codebook object mapping codewords to the targets they are designed to detect
        metric : str
            the sklearn metric string to pass to NearestNeighbors
        distance_threshold : float
            spots whose codewords are more than this metric distance from an expected code are filtered
        magnitude_threshold : int
            spots with intensity less than this value are filtered
        min_area : int
            spots with total area less than this value are filtered
        max_area : int
            spots with total area greater than this value are filtered
        norm_order : int
            order of L_p norm to apply to intensities and codes when using metric_decode to pair
            each intensities to its closest target (default = 2)
        crop_x, crop_y, crop_z : int
            number of pixels to crop from the top and bottom of each of the x, y, and z axes of
            an ImageStack (default = 0)

        """
        self.codebook = codebook
        self.metric = metric
        self.distance_threshold = distance_threshold
        self.magnitude_threshold = magnitude_threshold
        self.min_area = min_area
        self.max_area = max_area
        self.norm_order = norm_order
        self.crop_x = crop_x
        self.crop_y = crop_y
        self.crop_z = crop_z

    def find(self, stack: ImageStack) \
            -> Tuple[IntensityTable, ConnectedComponentDecodingResult]:
        """decode pixels and combine them into spots using connected component labeling

        Parameters
        ----------
        stack : ImageStack
            ImageStack containing spots

        Returns
        -------
        IntensityTable :
            IntensityTable containing decoded spots
        ConnectedComponentDecodingResult :
            Results of connected component labeling

        """
        pixel_intensities = IntensityTable.from_image_stack(
            stack, crop_x=self.crop_x, crop_y=self.crop_y, crop_z=self.crop_z)
        decoded_intensities = self.codebook.metric_decode(
            pixel_intensities,
            max_distance=self.distance_threshold,
            min_intensity=self.magnitude_threshold,
            norm_order=self.norm_order,
            metric=self.metric
        )
        decoded_spots, image_decoding_results = combine_adjacent_features(
            decoded_intensities,
            min_area=self.min_area,
            max_area=self.max_area
        )

        return decoded_spots, image_decoding_results

    @classmethod
    def add_arguments(cls, group_parser):
        group_parser.add_argument("--codebook-input", help="csv file containing a codebook")
        group_parser.add_argument("--metric", type='str', default='euclidean')
        group_parser.add_argument(
            "--distance-threshold", default=0.5176,
            help="maximum distance a pixel may be from a codeword before it is filtered")
        group_parser.add_argument("--magnitude-threshold", type=float, default=1, help="minimum magnitude of a feature")
        group_parser.add_argument("--area-threshold", type=float, default=2, help="minimum area of a feature")
        group_parser.add_argument('--crop-x', type=int, default=0)
        group_parser.add_argument('--crop-y', type=int, default=0)
        group_parser.add_argument('--crop-z', type=int, default=0)



from functools import partial
from typing import Optional

import numpy as np
from trackpy import bandpass

from starfish.image import ImageStack
from starfish.types import Number
from ._base import FilterAlgorithmBase


class Bandpass(FilterAlgorithmBase):

    def __init__(
            self, lshort: Number, llong: int, threshold: Number, truncate: Number=4,
            is_volume: bool=False, **kwargs) -> None:
        """

        Parameters
        ----------
        lshort : float
            filter frequencies below this value
        llong : int
            filter frequencies above this odd integer value
        threshold : float
            zero any pixels below this intensity value
        truncate : float
            # TODO dganguli: this is not documented by trackpy, can you help?
        is_volume : bool
            If True, 3d (z, y, x) volumes will be filtered. By default, filter 2-d (y, x) planes
        kwargs
        """
        self.lshort = lshort
        self.llong = llong
        self.threshold = threshold
        self.truncate = truncate
        self.is_volume = is_volume

    @classmethod
    def add_arguments(cls, group_parser) -> None:
        group_parser.add_argument(
            "--lshort", type=float, help="filter signals below this frequency")
        group_parser.add_argument(
            "--llong", type=int, help="filter signals above this frequency")
        group_parser.add_argument(
            "--threshold", type=int, help="clip pixels below this intensity value")
        group_parser.add_argument("--truncate", default=4, type=int)  # TODO dganguli doc this too

    @staticmethod
    def bandpass(
            image: np.ndarray, lshort: Number, llong: int, threshold: Number, truncate: Number
    ) -> np.ndarray:
        """Apply a bandpass filter to remove noise and background variation

        Parameters
        ----------
        image : np.ndarray
        lshort : float
            filter frequencies below this value
        llong : int
            filter frequencies above this odd integer value
        threshold : float
            zero any pixels below this intensity value
        truncate : float
             # TODO dganguli: this is not documented by trackpy, can you help?

        Returns
        -------
        np.ndarray :
            bandpassed image

        """
        bandpassed: np.ndarray = bandpass(
            image, lshort=lshort, llong=llong, threshold=threshold,
            truncate=truncate
        )
        return bandpassed

    def run(
            self, stack: ImageStack, in_place: bool=True, verbose: bool=False
    ) -> Optional[ImageStack]:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report the filtering progress across the tiles or volumes of the ImageStack

        Returns
        -------
        Optional[ImageStack] :
            if in-place is False, return the results of filter as a new stack

        """
        bandpass_ = partial(
            self.bandpass,
            lshort=self.lshort, llong=self.llong, threshold=self.threshold, truncate=self.truncate
        )
        result = stack.apply(
            bandpass_, verbose=verbose, in_place=in_place, is_volume=self.is_volume
        )
        if not in_place:
            return result
        return None



import json
import os
import shutil
import subprocess
import sys
import tempfile
import unittest
from typing import Sequence

import jsonpath_rw
import numpy as np
import pandas as pd

from starfish.constants import Features
from starfish.intensity_table import IntensityTable
from starfish.util import clock


def get_jsonpath_from_file(json_filepath_components: Sequence[str], jsonpath: str):
    """
    Given a series of filepath components, join them to find a file <FILE>.  Open that file, and locate a specific value
    in the json structure <PATH>.  Join the directory path of <FILE> and <PATH> and return that.

    For example, if json_filepath_components is ["/tmp", "formatted", "experiment.json"] and jsonpath is
    "$['hybridization']", this method will open /tmp/formatted/experiment.json, decode that as a json document, and
    locate the value of the key 'hybridization'.  It will return /tmp/formatted/XXX, where XXX is the value of the key.
    """
    json_filepath = os.path.join(*json_filepath_components)
    dirname = os.path.dirname(json_filepath)
    with open(json_filepath, "r") as fh:
        document = json.load(fh)
        return os.path.join(dirname, jsonpath_rw.parse(jsonpath).find(document)[0].value)


class TestWithIssData(unittest.TestCase):
    SUBDIRS = (
        "raw",
        "formatted",
        "registered",
        "filtered",
        "results",
    )

    STAGES = (
        [
            sys.executable,
            "examples/get_iss_data.py",
            lambda tempdir, *args, **kwargs: os.path.join(tempdir, "raw"),
            lambda tempdir, *args, **kwargs: os.path.join(tempdir, "formatted"),
            "--d", "1",
        ],
        [
            "starfish", "registration",
            "--input", lambda tempdir, *args, **kwargs: get_jsonpath_from_file(
                [tempdir, "formatted", "experiment.json"],
                "$['hybridization_images']",
            ),
            "--output", lambda tempdir, *args, **kwargs: os.path.join(tempdir, "registered", "hybridization.json"),
            "FourierShiftRegistration",
            "--reference-stack", lambda tempdir, *args, **kwargs: get_jsonpath_from_file(
                [tempdir, "formatted", "experiment.json"],
                "$['auxiliary_images']['dots']",
            ),
            "--upsampling", "1000",
        ],
        [
            "starfish", "filter",
            "--input", lambda tempdir, *args, **kwargs: os.path.join(tempdir, "registered", "hybridization.json"),
            "--output", lambda tempdir, *args, **kwargs: os.path.join(tempdir, "filtered", "hybridization.json"),
            "WhiteTophat",
            "--masking-radius", "15",
        ],
        [
            "starfish", "filter",
            "--input", lambda tempdir, *args, **kwargs: get_jsonpath_from_file(
                [tempdir, "formatted", "experiment.json"],
                "$['auxiliary_images']['nuclei']",
            ),
            "--output", lambda tempdir, *args, **kwargs: os.path.join(tempdir, "filtered", "nuclei.json"),
            "WhiteTophat",
            "--masking-radius", "15",
        ],
        [
            "starfish", "filter",
            "--input", lambda tempdir, *args, **kwargs: get_jsonpath_from_file(
                [tempdir, "formatted", "experiment.json"],
                "$['auxiliary_images']['dots']",
            ),
            "--output", lambda tempdir, *args, **kwargs: os.path.join(tempdir, "filtered", "dots.json"),
            "WhiteTophat",
            "--masking-radius", "15",
        ],
        [
            "starfish", "detect_spots",
            "--input", lambda tempdir, *args, **kwargs: os.path.join(tempdir, "filtered", "hybridization.json"),
            "--output", lambda tempdir, *args, **kwargs: os.path.join(tempdir, "results"),
            "GaussianSpotDetector",
            "--blobs-stack", lambda tempdir, *args, **kwargs: os.path.join(tempdir, "filtered", "dots.json"),
            "--min-sigma", "4",
            "--max-sigma", "6",
            "--num-sigma", "20",
            "--threshold", "0.01",
        ],
        [
            "starfish", "segment",
            "--hybridization-stack", lambda tempdir, *args, **kwargs: os.path.join(
                tempdir, "filtered", "hybridization.json"),
            "--nuclei-stack", lambda tempdir, *args, **kwargs: os.path.join(tempdir, "filtered", "nuclei.json"),
            "-o", lambda tempdir, *args, **kwargs: os.path.join(tempdir, "results", "regions.geojson"),
            "Watershed",
            "--dapi-threshold", ".16",
            "--input-threshold", ".22",
            "--min-distance", "57",
        ],
        [
            "starfish", "target_assignment",
            "--coordinates-geojson",
            lambda tempdir, *args, **kwargs: os.path.join(tempdir, "results", "regions.geojson"),
            "--intensities", lambda tempdir, *args, **kwargs: os.path.join(tempdir, "results", "spots.nc"),
            "--output", lambda tempdir, *args, **kwargs: os.path.join(tempdir, "results", "regions.json"),
            "PointInPoly2D",
        ],
        [
            "starfish", "decode",
            "-i", lambda tempdir, *args, **kwargs: os.path.join(tempdir, "results", "spots.nc"),
            "--codebook", lambda tempdir, *args, **kwargs: get_jsonpath_from_file(
                [tempdir, "formatted", "experiment.json"],
                "$['codebook']",
            ),
            "-o", lambda tempdir, *args, **kwargs: os.path.join(tempdir, "results", "spots.nc"),
            "IssDecoder",
        ],
    )

    def test_run_pipeline(self):
        tempdir = tempfile.mkdtemp()
        coverage_enabled = "STARFISH_COVERAGE" in os.environ

        def callback(interval):
            print(" ".join(stage[:2]), " ==> {} seconds".format(interval))

        try:
            for subdir in TestWithIssData.SUBDIRS:
                os.makedirs("{tempdir}".format(
                    tempdir=os.path.join(tempdir, subdir)))
            for stage in TestWithIssData.STAGES:
                cmdline = [
                    element(tempdir=tempdir) if callable(element) else element
                    for element in stage
                ]
                if cmdline[0] == "starfish" and coverage_enabled:
                    coverage_cmdline = [
                        "coverage", "run",
                        "-p",
                        "--source", "starfish",
                        "-m", "starfish",
                    ]
                    coverage_cmdline.extend(cmdline[1:])
                    cmdline = coverage_cmdline
                with clock.timeit(callback):
                    subprocess.check_call(cmdline)

            intensities = IntensityTable.load(os.path.join(tempdir, "results", "spots.nc"))
            genes, counts = np.unique(
                intensities.coords[Features.TARGET], return_counts=True)
            gene_counts = pd.Series(counts, genes)
            assert gene_counts['ACTB_human'] > gene_counts['ACTB_mouse']

        finally:
            if os.getenv("TEST_ISS_KEEP_DATA") is None:
                shutil.rmtree(tempdir)



import argparse
from functools import partial
from typing import Callable, Optional

import numpy as np
from skimage import restoration

from starfish.image import ImageStack
from ._base import FilterAlgorithmBase
from .util import gaussian_kernel
from starfish.types import Number


class DeconvolvePSF(FilterAlgorithmBase):

    def __init__(
            self, num_iter: int, sigma: Number, clip: bool=False, **kwargs) -> None:
        """Deconvolve a point spread function

        Parameters
        ----------
        num_iter : int
            number of iterations to run. Note that this is a very important parameter that requires
            careful optimization
        sigma : Number
            standard deviation of the gaussian kernel used to construct the point spread function
        clip : bool (default = False)
            if True, pixel values below -1 and above 1 are clipped for skimage pipeline
            compatibility

        """
        self.num_iter = num_iter
        self.sigma = sigma
        self.clip = clip
        self.kernel_size: int = int(2 * np.ceil(2 * sigma) + 1)
        self.psf: np.ndarray = gaussian_kernel(
            shape=(self.kernel_size, self.kernel_size),
            sigma=sigma
        )

    @classmethod
    def add_arguments(cls, group_parser: argparse.ArgumentParser) -> None:
        group_parser.add_argument(
            '--num-iter', type=int, help='number of iterations to run')
        group_parser.add_argument(
            '--sigma', type=float, help='standard deviation of gaussian kernel')
        group_parser.add_argument(
            '--clip', action='store_true',
            help='(default False) if True, clip values below -1 and above 1')

    @staticmethod
    def richardson_lucy_deconv(
            img: np.ndarray, num_iter: int, psf: np.ndarray, clip: bool=False) -> np.ndarray:
        """
        Deconvolves input image with a specified point spread function.

       Parameters
        ----------
        img : np.ndarray
            Image to filter.
        num_iter : int
            Number of iterations to run algorithm
        psf : np.ndarray
            Point spread function
        clip : bool (default = False)
            If true, pixel value of the result above 1 or under -1 are thresholded for skimage
            pipeline compatibility.

        Notes
        ------
        wrapper for skimage.restoration.richardson_lucy

        Returns
        -------
        np.ndarray :
            Deconvolved image, same shape as input

        """

        # TODO ambrosejcarr: the restoration function is producing the following warning:
        # /usr/local/lib/python3.6/site-packages/skimage/restoration/deconvolution.py:389:
        # RuntimeWarning: invalid value encountered in true_divide:
        # relative_blur = image / convolve_method(im_deconv, psf, 'same')
        img_deconv: np.ndarray = restoration.richardson_lucy(
            img, psf, iterations=num_iter, clip=clip
        )

        # here be dragons. img_deconv is a float. this should not work, but the result looks nice
        # modulo boundary values? wtf indeed.
        img_deconv = img_deconv.astype(np.uint16)
        return img_deconv

    def run(self, stack: ImageStack, in_place: bool=True, verbose=False) -> Optional[ImageStack]:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on the percentage completed during processing (default = False)

        Returns
        -------
        Optional[ImageStack] :
            if in-place is False, return the results of filter as a new stack

        """
        func: Callable = partial(self.richardson_lucy_deconv, num_iter=self.num_iter, psf=self.psf,
                                 clip=self.clip)
        result = stack.apply(func, in_place=in_place, verbose=verbose)
        if not in_place:
            return result
        return None



from functools import reduce
from typing import Dict, List, Sequence

import numpy as np
import pandas as pd
import scipy.ndimage.measurements as spm
import regional
from scipy.sparse import coo_matrix

from starfish.image import ImageStack
from starfish.types import Number


def stack_describe(stack: np.ndarray) -> List[Dict[str, Number]]:
    num_rounds = stack.shape[0]
    stats = [im_describe(stack[k, :]) for k in range(num_rounds)]
    return stats


def im_describe(im: np.ndarray) -> Dict[str, Number]:
    shape = im.shape
    flat_dims = reduce(lambda x, y: x * y, shape)
    flat_im = np.reshape(im, flat_dims)
    stats = pd.Series(flat_im).describe()
    return stats.to_dict()


def label_to_regions(labels) -> regional.many:
    label_mat_coo = coo_matrix(labels)

    def region_for(label_mat_coo, label):
        ind = label_mat_coo.data == label
        # TODO does this work in 3D?
        x = label_mat_coo.row[ind]
        y = label_mat_coo.col[ind]

        re = regional.one(list(zip(x, y)))
        return re

    unique_labels = sorted(set(label_mat_coo.data))
    regions = [region_for(label_mat_coo, label) for label in unique_labels]

    return regional.many(regions)


def measure(
        im: np.ndarray, labels: Sequence[Number], num_objs: int, measurement_type: str='mean'
) -> List[float]:
    if measurement_type == 'mean':
        res = spm.mean(im, labels, range(1, num_objs))
    elif measurement_type == 'max':
        res = spm.maximum(im, labels, range(1, num_objs))
    else:
        raise ValueError(f'Unsupported measurement type: {measurement_type}. Choose one of "mean" '
                         f'or "max".')

    return res


def measure_stack(
        stack: np.ndarray, labels: Sequence[Number], num_objs: int, measurement_type='mean'
) -> List[List[float]]:
    from starfish.munge import stack_to_list
    ims = stack_to_list(stack)
    res = [measure(im, labels, num_objs, measurement_type) for im in ims]
    return res



from typing import Tuple, Union

import numpy as np
import pytest

from starfish.io import Stack
from starfish.test.dataset_fixtures import merfish_stack
from starfish.pipeline.filter import gaussian_high_pass
from starfish.types import Number


@pytest.mark.parametrize('sigma', (1, (1, 1)))
def test_gaussian_high_pass(merfish_stack: Stack, sigma: Union[Number, Tuple[Number]]):
    """high pass is subtractive, sum of array should be less after running

    Parameters
    ----------
    merfish_stack : Stack
        pytest fixture that exposes a starfish.io.Stack object containing MERFISH testing data
    sigma : Union[int, Tuple[int]]
        the standard deviation of the gaussian kernel
    """
    ghp = gaussian_high_pass.GaussianHighPass(sigma=sigma)
    sum_before = np.sum(merfish_stack.image.numpy_array)
    ghp.run(merfish_stack.image)
    assert np.sum(merfish_stack.image.numpy_array) < sum_before


@pytest.mark.parametrize('sigma', (1, (1, 1)))
def test_gaussian_high_pass_apply(merfish_stack: Stack, sigma: Union[Number, Tuple[Number]]):
    """same as test_gaussian_high_pass, but tests apply loop functionality"""
    sum_before = np.sum(merfish_stack.image.numpy_array)
    ghp = gaussian_high_pass.GaussianHighPass(sigma=sigma)
    ghp.run(merfish_stack.image)
    assert np.sum(merfish_stack.image.numpy_array) < sum_before


@pytest.mark.parametrize('sigma', (1, (1, 0, 1)))
def test_gaussian_high_pass_apply_3d(merfish_stack: Stack, sigma: Union[Number, Tuple[Number]]):
    """same as test_gaussian_high_pass, but tests apply loop functionality"""
    sum_before = np.sum(merfish_stack.image.numpy_array)
    ghp = gaussian_high_pass.GaussianHighPass(sigma=sigma, is_volume=True)
    ghp.run(merfish_stack.image)
    assert np.sum(merfish_stack.image.numpy_array) < sum_before



import numpy as np
from skimage.filters import gaussian
import pytest

from starfish.pipeline.filter.white_tophat import WhiteTophat


def simple_spot_3d():

    big_spot = np.zeros((100, 100, 100), dtype=np.uint16)
    big_spot[20, 20, 20] = 10000
    big_spot = gaussian(big_spot, sigma=(4, 4, 4), preserve_range=True).astype(np.uint16)

    small_spot = np.zeros((100, 100, 100), dtype=np.uint16)
    small_spot[80, 80, 80] = 10000
    small_spot = gaussian(small_spot, sigma=(1, 1, 1), preserve_range=True).astype(np.uint16)

    return big_spot + small_spot


@pytest.mark.parametrize('is_volume', [True, False])
def test_white_tophat(is_volume: bool):
    """
    white tophat filters should reduce the intensity of spots that are larger than the masking
    radius much more than those that are smaller. Here we have a 4-diameter spot and a 2-diameter
    spot. We expect the large spot to be filtered more substantially.

    Verify this functionality in both 2d and 3d images
    """
    if is_volume:

        image = simple_spot_3d()
        small_spot_intensity = image[80, 80, 80]
        big_spot_intensity = image[20, 20, 20]

        wth = WhiteTophat(masking_radius=2, is_volume=True)
        filtered = wth.white_tophat(image)

        large_ratio = filtered[20, 20, 20] / big_spot_intensity
        small_ratio = filtered[80, 80, 80] / small_spot_intensity

    else:

        image = simple_spot_3d().max(axis=0)
        small_spot_intensity = image[80, 80]
        big_spot_intensity = image[20, 20]

        wth = WhiteTophat(masking_radius=2, is_volume=False)
        filtered = wth.white_tophat(image)

        large_ratio = filtered[20, 20] / big_spot_intensity
        small_ratio = filtered[80, 80] / small_spot_intensity

    assert large_ratio < small_ratio



import pytest

from starfish.constants import Indices
from starfish.test.dataset_fixtures import merfish_stack
from starfish.pipeline.features.pixels.pixel_spot_detector import PixelSpotDetector
from starfish.pipeline.filter.gaussian_high_pass import GaussianHighPass
from starfish.pipeline.filter.gaussian_low_pass import GaussianLowPass
from starfish.pipeline.filter.richardson_lucy_deconvolution import DeconvolvePSF


@pytest.mark.skip("TODO ambrosejcarr: fix this test in a future PR. Currently no spots are generated, and this fails.")
def test_merfish_pipeline(merfish_stack):
    s = merfish_stack

    # high pass filter
    ghp = GaussianHighPass(sigma=3)
    ghp.run(s)

    # deconvolve the point spread function
    dpsf = DeconvolvePSF(num_iter=15, sigma=2)
    dpsf.run(s)

    # low pass filter
    glp = GaussianLowPass(sigma=1)
    glp.run(s)

    # scale the data by the scale factors
    scale_factors = {(t[Indices.ROUND], t[Indices.CH]): t['scale_factor'] for index, t in s.tile_metadata.iterrows()}
    for indices in s.image._iter_indices():
        data = s.image.get_slice(indices)[0]
        scaled = data / scale_factors[indices[Indices.ROUND], indices[Indices.CH]]
        s.image.set_slice(indices, scaled)

    # detect and decode spots
    psd = PixelSpotDetector(
        codebook='https://s3.amazonaws.com/czi.starfish.data.public/MERFISH/codebook.csv',
        metric='euclidean',
        distance_threshold=0.5176,
        magnitude_threshold=1,
        area_threshold=2,
        crop_size=40
    )

    psd.find(s)



import numpy as np
import pandas as pd

from starfish.munge import dataframe_to_multiindex


def test_dataframe_from_multiindex():
    data = np.arange(4).reshape(2, 2)
    names = list('cd')
    df = pd.DataFrame(data, index=list('ab'), columns=names)
    mi = dataframe_to_multiindex(df)

    assert mi.names == names
    assert len(mi.levels) == 2
    assert np.array_equal(mi.get_level_values('c'), np.array([0, 2]))
    assert np.array_equal(mi.get_level_values('d'), np.array([1, 3]))



import numpy as np
import pytest

from starfish.constants import Indices, Features
from starfish.image import ImageStack
from starfish.intensity_table import IntensityTable
# don't inspect pytest fixtures in pycharm
# noinspection PyUnresolvedReferences
from starfish.test.dataset_fixtures import (
    synthetic_intensity_table,
    synthetic_spot_pass_through_stack,
    loaded_codebook,
    synthetic_dataset_with_truth_values,
    simple_codebook_json,
    simple_codebook_array,
    single_synthetic_spot
)


def test_get_slice_simple_index():
    """
    Retrieve a slice across one of the indices at the end.  For instance, if the dimensions are
    (P, Q0,..., Qn-1, R), slice across either P or R.
    """
    stack = ImageStack.synthetic_stack()
    round_ = 1
    imageslice, axes = stack.get_slice(
        {Indices.ROUND: round_}
    )
    assert axes == [Indices.CH, Indices.Z]

    y, x = stack.tile_shape

    for ch in range(stack.shape[Indices.CH]):
        for z in range(stack.shape[Indices.Z]):
            data = np.empty((y, x))
            data.fill((round_ * stack.shape[Indices.CH] + ch) * stack.shape[Indices.Z] + z)

            assert data.all() == imageslice[ch, z].all()


def test_get_slice_middle_index():
    """
    Retrieve a slice across one of the indices in the middle.  For instance, if the dimensions are
    (P, Q0,..., Qn-1, R), slice across one of the Q axes.
    """
    stack = ImageStack.synthetic_stack()
    ch = 1
    imageslice, axes = stack.get_slice(
        {Indices.CH: ch}
    )
    assert axes == [Indices.ROUND, Indices.Z]

    y, x = stack.tile_shape

    for round_ in range(stack.shape[Indices.ROUND]):
        for z in range(stack.shape[Indices.Z]):
            data = np.empty((y, x))
            data.fill((round_ * stack.shape[Indices.CH] + ch) * stack.shape[Indices.Z] + z)

            assert data.all() == imageslice[round_, z].all()


def test_get_slice_range():
    """
    Retrieve a slice across a range of one of the dimensions.
    """
    stack = ImageStack.synthetic_stack()
    zrange = slice(1, 3)
    imageslice, axes = stack.get_slice(
        {Indices.Z: zrange}
    )
    y, x = stack.tile_shape
    assert axes == [Indices.ROUND, Indices.CH, Indices.Z]

    for round_ in range(stack.shape[Indices.ROUND]):
        for ch in range(stack.shape[Indices.CH]):
            for z in range(zrange.stop - zrange.start):
                data = np.empty((y, x))
                data.fill((round_ * stack.shape[Indices.CH] + ch) * stack.shape[Indices.Z] +
                          (z + zrange.start))

                assert data.all() == imageslice[round_, ch, z].all()


def test_set_slice_simple_index():
    """
    Sets a slice across one of the indices at the end.  For instance, if the dimensions are
    (P, Q0,..., Qn-1, R), sets a slice across either P or R.
    """
    stack = ImageStack.synthetic_stack()
    round_ = 1
    y, x = stack.tile_shape

    expected = np.ones((stack.shape[Indices.CH], stack.shape[Indices.Z], y, x)) * 2
    index = {Indices.ROUND: round_}

    stack.set_slice(index, expected)

    assert np.array_equal(stack.get_slice(index)[0], expected)


def test_set_slice_middle_index():
    """
    Sets a slice across one of the indices in the middle.  For instance, if the dimensions are
    (P, Q0,..., Qn-1, R), slice across one of the Q axes.
    """
    stack = ImageStack.synthetic_stack()
    ch = 1
    y, x = stack.tile_shape

    expected = np.ones((stack.shape[Indices.ROUND], stack.shape[Indices.Z], y, x)) * 2
    index = {Indices.CH: ch}

    stack.set_slice(index, expected)

    assert np.array_equal(stack.get_slice(index)[0], expected)


def test_set_slice_range():
    """
    Sets a slice across a range of one of the dimensions.
    """
    stack = ImageStack.synthetic_stack()
    zrange = slice(1, 3)
    y, x = stack.tile_shape

    expected = np.ones((
        stack.shape[Indices.ROUND],
        stack.shape[Indices.CH],
        zrange.stop - zrange.start,
        y, x)) * 10
    index = {Indices.Z: zrange}

    stack.set_slice(index, expected)

    assert np.array_equal(stack.get_slice(index)[0], expected)


def test_from_numpy_array_preserves_data():
    array = np.random.random((1, 1, 1, 2, 2))
    image_stack = ImageStack.from_numpy_array(array)
    assert np.array_equal(array, image_stack.numpy_array)


def test_from_numpy_array_raises_error_when_incorrect_dims_passed():
    array = np.ones((2, 2))
    # verify this method works with the correct shape
    image = ImageStack.from_numpy_array(array.reshape((1, 1, 1, 2, 2)))
    assert isinstance(image, ImageStack)

    with pytest.raises(ValueError):
        ImageStack.from_numpy_array(array.reshape((1, 1, 2, 2)))
        ImageStack.from_numpy_array(array.reshape((1, 2, 2)))
        ImageStack.from_numpy_array(array)
        ImageStack.from_numpy_array(array.reshape((1, 1, 1, 1, 2, 2)))


def test_from_numpy_array_preserves_dtype():
    original_dtype = np.uint16
    array = np.ones((2, 2, 2), dtype=original_dtype)
    image = ImageStack.from_numpy_array(array.reshape((1, 1, 2, 2, 2)))
    assert image.numpy_array.dtype == original_dtype


def test_max_projection_preserves_dtype():
    original_dtype = np.uint16
    array = np.ones((2, 2, 2), dtype=original_dtype)
    image = ImageStack.from_numpy_array(array.reshape((1, 1, 2, 2, 2)))

    max_projection = image.max_proj(Indices.CH, Indices.ROUND, Indices.Z)
    assert max_projection.dtype == original_dtype


def test_synthetic_spot_creation_raises_error_with_coords_too_small(synthetic_intensity_table):
    num_z = 0
    height = 40
    width = 50
    with pytest.raises(ValueError):
        ImageStack.synthetic_spots(synthetic_intensity_table, num_z, height, width)


def test_synthetic_spot_creation_produces_an_imagestack(synthetic_intensity_table):
    num_z = 12
    height = 50
    width = 40
    image = ImageStack.synthetic_spots(synthetic_intensity_table, num_z, height, width)
    assert isinstance(image, ImageStack)


def test_synthetic_spot_creation_produces_an_imagestack_with_correct_spot_location(
        synthetic_spot_pass_through_stack):

    codebook, true_intensities, image = synthetic_spot_pass_through_stack

    g, c, h = np.where(true_intensities.values)

    x = np.empty_like(g)
    y = np.empty_like(g)
    z = np.empty_like(g)
    breaks = np.concatenate([
        np.array([0]),
        np.where(np.diff(g))[0] + 1,
        np.array([g.shape[0]])
    ])
    for i in np.arange(len(breaks) - 1):
        x[breaks[i]: breaks[i + 1]] = true_intensities.coords[Features.X][i]
        y[breaks[i]: breaks[i + 1]] = true_intensities.coords[Features.Y][i]
        z[breaks[i]: breaks[i + 1]] = true_intensities.coords[Features.Z][i]

    # only 8 values should be set, since there are only 8 locations across the tensor
    assert np.sum(image.numpy_array != 0) == 8

    assert np.array_equal(
        image.numpy_array[h, c, z, y, x],
        true_intensities.values[np.where(true_intensities)])


# TODO ambrosejcarr: improve the tests here.
def test_imagestack_to_intensity_table(single_synthetic_spot):
    codebook, intensity_table, image = single_synthetic_spot
    pixel_intensities = IntensityTable.from_image_stack(image)
    pixel_intensities = codebook.metric_decode(
        pixel_intensities, max_distance=0, min_intensity=1000, norm_order=2)
    assert isinstance(pixel_intensities, IntensityTable)


def test_imagestack_to_intensity_table_no_noise(synthetic_spot_pass_through_stack):
    codebook, intensity_table, image = synthetic_spot_pass_through_stack
    pixel_intensities = IntensityTable.from_image_stack(image)
    pixel_intensities = codebook.metric_decode(
        pixel_intensities, max_distance=0, min_intensity=1000, norm_order=2)
    assert isinstance(pixel_intensities, IntensityTable)



#!/usr/bin/env python

import os
import setuptools

CLASSIFIERS = [
    "Development Status :: 2 - Pre-Alpha",
    "Natural Language :: English",
    "Operating System :: POSIX",
    "Operating System :: Unix",
    "Operating System :: MacOS :: MacOS X",
    "License :: OSI Approved :: MIT License",
    "Programming Language :: Python :: 3.6",
    "Topic :: Scientific/Engineering :: Bio-Informatics",
]

install_requires = [line.rstrip() for line in open(os.path.join(os.path.dirname(__file__), "REQUIREMENTS.txt"))]

setuptools.setup(
    name="starfish",
    version="0.0.9",
    description="Pipelines and pipeline components for the analysis of image-based transcriptomics data",
    author="Deep Ganguli",
    author_email="dganguli@chanzuckerberg.com",
    license="MIT",
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    entry_points={
        'console_scripts': "starfish=starfish.starfish:starfish"
    },
    classifiers=CLASSIFIERS
)



from typing import Any

import pytest

from starfish.constants import Indices
from starfish.image import ImageStack

NUM_ROUND = 4
NUM_CH = 2
NUM_Z = 12


def test_metadata():
    """
    Normal situation where all the tiles have uniform keys for both indices and extras.
    """
    def tile_extras_provider(round_: int, ch: int, z: int) -> Any:
        return {
            'random_key': {
                Indices.ROUND: round_,
                Indices.CH: ch,
                Indices.Z: z,
            }
        }

    stack = ImageStack.synthetic_stack(
        num_round=NUM_ROUND, num_ch=NUM_CH, num_z=NUM_Z, tile_extras_provider=tile_extras_provider,
    )
    table = stack.tile_metadata
    assert len(table) == NUM_ROUND * NUM_CH * NUM_Z


def test_missing_extras():
    """
    If the extras are not present on some of the tiles, it should still work.
    """
    def tile_extras_provider(round_: int, ch: int, z: int) -> Any:
        if round_ == 0:
            return {
                'random_key': {
                    Indices.ROUND: round_,
                    Indices.CH: ch,
                    Indices.Z: z,
                }
            }
        else:
            return None

    stack = ImageStack.synthetic_stack(
        num_round=NUM_ROUND, num_ch=NUM_CH, num_z=NUM_Z, tile_extras_provider=tile_extras_provider,
    )
    table = stack.tile_metadata
    assert len(table) == NUM_ROUND * NUM_CH * NUM_Z


def test_conflict():
    """
    Tiles that have extras that conflict with indices should produce an error.
    """
    def tile_extras_provider(round_: int, ch: int, z: int) -> Any:
        return {
            Indices.ROUND: round_,
            Indices.CH: ch,
            Indices.Z: z,
        }

    stack = ImageStack.synthetic_stack(
        num_round=NUM_ROUND, num_ch=NUM_CH, num_z=NUM_Z, tile_extras_provider=tile_extras_provider,
    )
    with pytest.raises(ValueError):
        stack.tile_metadata



import pandas as pd
import regional

from starfish.intensity_table import IntensityTable
from starfish.pipeline.algorithmbase import AlgorithmBase


class TargetAssignmentAlgorithm(AlgorithmBase):
    def run(self, spots: IntensityTable, regions: regional.many) -> pd.DataFrame:
        """Performs target (e.g. gene) assignment given the spots and the regions."""
        raise NotImplementedError()



from itertools import product
from typing import Union, Tuple

import numpy as np
import pandas as pd
import xarray as xr

from starfish.constants import Indices, Features
from starfish.munge import dataframe_to_multiindex


class IntensityTable(xr.DataArray):
    """3 dimensional container for spot/pixel features extracted from image data

    An IntensityTable is comprised of each feature's intensity across channels and imaging
    rounds, where features are typically spots or pixels. This forms an
    (n_feature, n_channel, n_round) tensor implemented as an xarray.DataArray object.
    In addition to the basic xarray methods, IntensityTable implements:

    Constructors
    -------
    empty_intensity_table  creates an IntensityTable with all intensities equal to zero
    from_spot_data         creates an IntensityTable from a 3d array and a spot attributes dataframe
    synthetic_intensities  creates an IntensityTable with synthetic spots, given a codebook

    Methods
    -------
    save                   save the IntensityTable to netCDF
    load                   load an IntensityTable from netCDF

    Examples
    --------
    >>> from starfish.util.synthesize import SyntheticData
    >>> sd = SyntheticData(n_ch=3, n_round=4, n_codes=2)
    >>> codes = sd.codebook()
    >>> sd.intensities(codebook=codes)
    <xarray.IntensityTable (features: 2, c: 3, h: 4)>
    array([[[    0.,     0.,     0.,     0.],
            [    0.,     0.,  8022., 12412.],
            [11160.,  9546.,     0.,     0.]],

           [[    0.,     0.,     0.,     0.],
            [    0.,     0., 10506., 10830.],
            [11172., 12331.,     0.,     0.]]])
    Coordinates:
    * features   (features) MultiIndex
    - z          (features) int64 7 3
    - y          (features) int64 14 32
    - x          (features) int64 32 15
    - r          (features) float64 nan nan
    * c          (c) int64 0 1 2
    * h          (h) int64 0 1 2 3
      target     (features) object 08b1a822-a1b4-4e06-81ea-8a4bd2b004a9 ...

    """

    # constant that stores the attr key for the shape of ImageStack
    IMAGE_SHAPE = 'image_shape'

    @classmethod
    def empty_intensity_table(
            cls, spot_attributes: pd.MultiIndex, n_ch: int, n_round: int,
            image_shape: Tuple[int, int, int]
    ) -> "IntensityTable":
        """Create an empty intensity table with pre-set axis whose values are zero

        Parameters
        ----------
        spot_attributes : pd.MultiIndex
            MultiIndex containing spot metadata. Must contain the values specifid in Constants.X,
            Y, Z, and RADIUS.
        n_ch : int
            number of channels measured in the imaging experiment
        n_round : int
            number of imaging rounds measured in the imaging experiment
        image_shape : Tuple[int, int, int]
            the shape (z, y, x) of the image from which features will be extracted

        Returns
        -------
        IntensityTable :
            empty IntensityTable

        """
        cls._verify_spot_attributes(spot_attributes)
        channel_index = np.arange(n_ch)
        round_index = np.arange(n_round)
        data = np.zeros((spot_attributes.shape[0], n_ch, n_round))
        dims = (Features.AXIS, Indices.CH.value, Indices.ROUND.value)
        attrs = {cls.IMAGE_SHAPE: image_shape}

        intensity_table = cls(
            data=data, coords=(spot_attributes, channel_index, round_index), dims=dims,
            attrs=attrs
        )

        return intensity_table

    @staticmethod
    def _verify_spot_attributes(spot_attributes: pd.MultiIndex) -> None:
        """Run some checks on spot attributes"""
        if not isinstance(spot_attributes, pd.MultiIndex):
            raise ValueError(
                f'spot attributes must be a pandas MultiIndex, not {type(spot_attributes)}.')

        required_attributes = {Features.Z, Features.Y, Features.X}
        missing_attributes = required_attributes.difference(spot_attributes.names)
        if missing_attributes:
            raise ValueError(
                f'Missing spot_attribute levels in provided MultiIndex: {missing_attributes}. '
                f'The following levels are required: {required_attributes}.')

    @classmethod
    def from_spot_data(
            cls, intensities: Union[xr.DataArray, np.ndarray], spot_attributes: pd.MultiIndex,
            image_shape: Tuple[int, int, int],
            *args, **kwargs) -> "IntensityTable":
        """Table to store image feature intensities and associated metadata

        Parameters
        ----------
        intensities : np.ndarray[Any]
            intensity data
        spot_attributes : pd.MultiIndex
            Name(s) of the data dimension(s). Must be either a string (only
            for 1D data) or a sequence of strings with length equal to the
            number of dimensions. If this argument is omitted, dimension names
            are taken from ``coords`` (if possible) and otherwise default to
            ``['dim_0', ... 'dim_n']``.
        image_shape : Tuple[int, int, int]
            the shape of the image (z, y, x) from which the features were extracted
        args :
            additional arguments to pass to the xarray constructor
        kwargs :
            additional keyword arguments to pass to the xarray constructor

        Returns
        -------
        IntensityTable :
            IntensityTable containing data from passed ndarray, annotated by spot_attributes

        """

        if len(intensities.shape) != 3:
            raise ValueError(
                f'intensities must be a (features * ch * round) 3-d tensor. Provided intensities '
                f'shape ({intensities.shape}) is invalid.')

        cls._verify_spot_attributes(spot_attributes)

        coords = (
            (Features.AXIS, spot_attributes),
            (Indices.CH.value, np.arange(intensities.shape[1])),
            (Indices.ROUND.value, np.arange(intensities.shape[2]))
        )

        dims = (Features.AXIS, Indices.CH.value, Indices.ROUND.value)

        attrs = {cls.IMAGE_SHAPE: image_shape}

        return cls(intensities, coords, dims, attrs=attrs, *args, **kwargs)

    def save(self, filename: str) -> None:
        """Save an IntensityTable as a Netcdf File

        Parameters
        ----------
        filename : str
            Name of Netcdf file

        """
        # TODO when https://github.com/pydata/xarray/issues/1077 (support for multiindex
        # serliazation) is merged, remove this reset_index() call and simplify load, below
        self.reset_index('features').to_netcdf(filename)

    @classmethod
    def load(cls, filename: str) -> "IntensityTable":
        """load an IntensityTable from Netcdf

        Parameters
        ----------
        filename : str
            File to load

        Returns
        -------
        IntensityTable

        """
        loaded = xr.open_dataarray(filename)
        intensity_table = cls(
            loaded.data,
            loaded.coords,
            loaded.dims
        )
        return intensity_table.set_index(
            features=list(intensity_table[Features.AXIS].coords.keys()))

    def show(self, background_image: np.ndarray) -> None:
        """show spots on a background image"""
        raise NotImplementedError

    @classmethod
    def synthetic_intensities(
            cls, codebook, num_z: int=12, height: int=50, width: int=40, n_spots=10,
            mean_fluor_per_spot=200, mean_photons_per_fluor=50
    ) -> "IntensityTable":
        """Create an IntensityTable containing synthetic spots with random locations

        Parameters
        ----------
        codebook : Codebook
            starfish codebook object
        num_z :
            number of z-planes to use when localizing spots
        height :
            y dimension of each synthetic plane
        width :
            x dimension of each synthetic plane
        n_spots :
            number of spots to generate
        mean_fluor_per_spot :
            mean number of fluorophores per spot
        mean_photons_per_fluor :
            mean number of photons per fluorophore.

        Returns
        -------
        IntensityTable

        """

        # TODO nsofroniew: right now there is no jitter on x-y positions of the spots
        z = np.random.randint(0, num_z, size=n_spots)
        y = np.random.randint(0, height, size=n_spots)
        x = np.random.randint(0, width, size=n_spots)
        r = np.empty(n_spots)
        r.fill(np.nan)  # radius is a function of the point-spread gaussian size

        names = [Features.Z, Features.Y, Features.X, Features.SPOT_RADIUS]
        spot_attributes = pd.MultiIndex.from_arrays([z, y, x, r], names=names)

        # empty data tensor
        data = np.zeros(shape=(n_spots, *codebook.shape[1:]))

        targets = np.random.choice(
            codebook.coords[Features.TARGET], size=n_spots, replace=True)
        expected_bright_locations = np.where(codebook.loc[targets])

        # create a binary matrix where "on" spots are 1
        data[expected_bright_locations] = 1

        # add physical properties of fluorescence
        data *= np.random.poisson(mean_photons_per_fluor, size=data.shape)
        data *= np.random.poisson(mean_fluor_per_spot, size=data.shape)

        image_shape = (num_z, height, width)

        intensities = cls.from_spot_data(data, spot_attributes, image_shape=image_shape)
        intensities[Features.TARGET] = (Features.AXIS, targets)

        return intensities

    @classmethod
    def from_image_stack(cls, image_stack, crop_x: int=0, crop_y: int=0, crop_z: int=0) -> "IntensityTable":
        """Generate an IntensityTable from all the pixels in the ImageStack

        Parameters
        ----------
        crop_x : int
            number of pixels to crop from both top and bottom of x
        crop_y : int
            number of pixels to crop from both top and bottom of y
        crop_z : int
            number of pixels to crop from both top and bottom of z
        image_stack : ImageStack
            ImageStack containing pixels to be treated as intensities

        Returns
        -------
        IntensityTable
            IntensityTable containing one intensity per pixel (across channels and rounds)

        """

        # verify the image is large enough to crop
        assert crop_z * 2 < image_stack.shape['z']
        assert crop_y * 2 < image_stack.shape['y']
        assert crop_x * 2 < image_stack.shape['x']

        zmin = crop_z
        ymin = crop_y
        xmin = crop_x
        zmax = image_stack.shape['z'] - crop_z
        ymax = image_stack.shape['y'] - crop_y
        xmax = image_stack.shape['x'] - crop_x
        data = image_stack.numpy_array.transpose(2, 3, 4, 1, 0)  # (z, y, x, ch, round)

        # crop and reshape imagestack to create IntensityTable data
        cropped_data = data[zmin:zmax, ymin:ymax, xmin:xmax, :, :]
        # (pixels, ch, round)
        intensity_data = cropped_data.reshape(-1, image_stack.num_chs, image_stack.num_rounds)

        # IntensityTable pixel coordinates
        z = np.arange(zmin, zmax)
        y = np.arange(ymin, ymax)
        x = np.arange(xmin, xmax)

        pixel_coordinates = pd.DataFrame(
            data=np.array(list(product(z, y, x))),
            columns=['z', 'y', 'x']
        )
        pixel_coordinates[Features.SPOT_RADIUS] = np.full(
            pixel_coordinates.shape[0], fill_value=np.nan)

        spot_attributes = dataframe_to_multiindex(pixel_coordinates)
        image_size = cropped_data.shape[:3]

        return IntensityTable.from_spot_data(intensity_data, spot_attributes, image_size)

    def mask_low_intensity_features(self, intensity_threshold):
        """return the indices of features that have average intensity below intensity_threshold"""
        mask = np.where(
            self.mean([Indices.CH.value, Indices.ROUND.value]).values < intensity_threshold)[0]
        return mask

    def mask_small_features(self, min_size: int, max_size: int):
        """return the indices of features whose radii are smaller than size_threshold"""
        mask = np.where(self.coords.features[Features.SPOT_RADIUS] < min_size)[0]
        mask |= np.where(self.coords.features[Features.SPOT_RADIUS] > max_size)[0]
        return mask

    def _intensities_from_regions(self, props, reduce_op='max') -> "IntensityTable":
        """turn regions back into intensities by reducing over the labeled area"""
        raise NotImplementedError



#!/usr/bin/env python

import argparse
import cProfile
from pstats import Stats

from .pipeline.features.spots.decoder import Decoder
from .pipeline.features.spots.detector import SpotFinder
from .pipeline.filter import Filter
from .pipeline.target_assignment import TargetAssignment
from .pipeline.registration import Registration
from .pipeline.segmentation import Segmentation
from .util.argparse import FsExistsType


def build_parser():
    parser = argparse.ArgumentParser()

    parser.add_argument("--profile", action="store_true", help="enable profiling")
    parser.add_argument("--noop", help=argparse.SUPPRESS, dest="starfish_command", action="store_const", const=noop)

    subparsers = parser.add_subparsers(dest="starfish_command")

    Registration.add_to_parser(subparsers)
    Filter.add_to_parser(subparsers)
    SpotFinder.add_to_parser(subparsers)
    Segmentation.add_to_parser(subparsers)
    TargetAssignment.add_to_parser(subparsers)
    Decoder.add_to_parser(subparsers)

    show_group = subparsers.add_parser("show")
    show_group.add_argument("in_json", type=FsExistsType())
    show_group.add_argument("--sz", default=10, type=int, help="Figure size")
    show_group.set_defaults(starfish_command=show)

    return parser


PROFILER_KEY = "profiler"
"""This is the dictionary key we use to attach the profiler to pass to the resultcallback."""
PROFILER_LINES = 15
"""This is the number of profiling rows to dump when --profile is enabled."""


def starfish():
    parser = build_parser()
    args, argv = parser.parse_known_args()

    art = """
         _              __ _     _
        | |            / _(_)   | |
     ___| |_ __ _ _ __| |_ _ ___| |__
    / __| __/ _` | '__|  _| / __| '_  `
    \__ \ || (_| | |  | | | \__ \ | | |
    |___/\__\__,_|_|  |_| |_|___/_| |_|

    """
    print(art)
    if args.profile:
        profiler = cProfile.Profile()
        profiler.enable()

    if args.starfish_command is None:
        parser.print_help()
        parser.exit(status=2)
    args.starfish_command(args, len(argv) != 0)

    if args.profile:
        stats = Stats(profiler)
        stats.sort_stats('tottime').print_stats(PROFILER_LINES)


def show(args, print_help=False):
    import matplotlib.pyplot as plt
    from showit import tile

    from .io import Stack

    s = Stack()
    s.read(args.in_json)
    tile(s.image.squeeze(), size=args.sz, bar=True)
    plt.show()


def noop(args, print_help=False):
    pass



import pytest

from starfish.io import Stack


def test_min_version():
    with pytest.raises(ValueError):
        Stack.verify_version("0.0.0-dev")
    Stack.verify_version(str(Stack.MIN_SUPPORTED_VERSION))
    Stack.verify_version(str(Stack.MAX_SUPPORTED_VERSION))


def test_max_version():
    with pytest.raises(ValueError):
        Stack.verify_version("4294967296.0.0")
    Stack.verify_version(str(Stack.MIN_SUPPORTED_VERSION))
    Stack.verify_version(str(Stack.MAX_SUPPORTED_VERSION))
